###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# General StarThinker constants that can also be used from the command line interface
from starthinker.config import UI_ROOT, UI_PROJECT, UI_CLIENT, UI_SERVICE, UI_CRON, UI_ZONE, DEVELOPMENT_MODE

# Load all technical settings
from starthinker_ui.ui.framework import *

# Used by workers to manage instance group
WORKER_MAX = int(os.environ.get('STARTHINKER_WORKER_MAX', 0))
WORKER_JOBS = int(os.environ.get('STARTHINKER_WORKER_JOBS', 1))

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = DEVELOPMENT_MODE

# Optional tracking parameter
GOOGLE_ANALYTICS = os.environ.get('STARTHINKER_ANALYTICS', '')

# Currently StarThinker is in English, internationalization TBD
LANGUAGE_CODE = 'en-us'

# Only affects logging on local server, all times in UI are recipe specific.
TIME_ZONE = 'America/Los_Angeles'

# If sendmail is set up, will email if site failure or error occurs.
ADMINS = [('User Name', 'email@domain.com')]

# Store your logo and front end website graphics here.
STATIC_URL = 'https://storage.googleapis.com/starthinker-ui/'


def domain_to_host(domain):
  return domain.split('://', 1)[-1].split(':', 1)[0].split('/', 1)[0]


CONST_URL = os.environ.get('STARTHINKER_UI_DOMAIN') or 'http://localhost:8000'
SECRET_KEY = os.environ.get(
    'STARTHINKER_UI_SECRET'
) or 'safetyandcivilreassuranceadministrationofficials'
ALLOWED_HOSTS = [
    domain_to_host(os.environ.get('STARTHINKER_UI_DOMAIN', '')) or 'localhost',
    '127.0.0.1'
]

DATABASES = {
    'default': {
        'ENGINE':
            os.environ.get('STARTHINKER_UI_DATABASE_ENGINE',
                           'MISSING RUN deploy.sh TO SET'),
        'HOST':
            os.environ.get('STARTHINKER_UI_DATABASE_HOST', ''),
        'PORT':
            os.environ.get('STARTHINKER_UI_DATABASE_PORT', ''),
        'NAME':
            os.environ.get('STARTHINKER_UI_DATABASE_NAME',
                           'MISSING RUN deploy.sh TO SET'),
        'USER':
            os.environ.get('STARTHINKER_UI_DATABASE_USER', ''),
        'PASSWORD':
            os.environ.get('STARTHINKER_UI_DATABASE_PASSWORD', ''),
        'TEST': {
            'NAME': 'starthinker_test'
        }
    }
}



import uuid
import os

from starthinker.util.project import project
from starthinker.util.google_api import API_DCM
from starthinker.util.bigquery import query_to_rows


@project.from_parameters
def publisher_preview():
  if project.verbose:
    print('Publisher Preview')

  profile_id = project.task['profile_id']

  #local_file = '/tmp/%s' % str(uuid.uuid1())
  local_file = '/tmp/%s' % 'test.html'

  out = open(local_file, 'wb')

  out.write('<html><body>')
  out.write('<table border=1>')
  out.write('<tr>')
  out.write(
      '<td><strong>Placement ID</strong></td><td><strong>Placement Name</strong></td><td><strong>Preview</strong></td>'
  )
  out.write('</tr>')

  placements = API_DCM(project.task['auth']).placements().list(
      profileId=profile_id,
      maxResults=5,
      campaignIds=[21097064, 21081031, 21108999, 21115284, 21072124,
                   21074779]).execute()

  for placement in placements['placements']:
    tag = API_DCM(project.task['auth']).placements().generatetags(
        profileId=profile_id,
        campaignId=placement['campaignId'],
        placementIds=[placement['id']],
        tagFormats=['PLACEMENT_TAG_IFRAME_JAVASCRIPT']).execute()

    for placement_tags in tag['placementTags']:
      for tag_datas in placement_tags['tagDatas']:
        out.write('<tr>')
        out.write('<td>%s</td>' % placement['id'])
        out.write('<td>%s</td>' % placement['name'])
        out.write('<td>%s</td>' % tag_datas['impressionTag'])
        out.write('</tr>')

  out.write('</table>')
  out.write('</body></html>')
  out.close()
  #os.path.remove(local_file)


if __name__ == '__main__':
  publisher_preview()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
import os

from starthinker.util.project import project
from starthinker.util.bigquery import query_to_table, query_to_rows, rows_to_table, table_create
from starthinker.util.csv import rows_to_type
from starthinker.util.sheets import sheets_read, sheets_clear, sheets_write
from starthinker.task.itp_audit.queries import Queries

# Output Tables
BROWSER_PERFORMANCE_2YEARS_TABLE = 'DV3_Browser'
SAFARI_DISTRIBUTION_90DAYS_TABLE = 'DV3_Safari'
DV360_BROWSER_SHARES_MULTICHART_TABLE = 'DV3_MultiChart'
CM_BROWSER_REPORT_CLEAN_TABLE = 'CM_Browser'
CM_FLOODLIGHT_MULTICHART_TABLE = 'CM_Floodlight_Multichart'

# Tables
CLEAN_BROWSER_REPORT_TABLE = 'z_DV360_Browser_Report_Clean'
DV360_CUSTOM_SEGMENTS_TABLE = 'z_Custom_Segments'
DV360_CUSTOM_SEGMENTS_SHEET_TABLE = 'z_Custom_Segments_Sheet'
CM_SITE_SEGMENTATION_SHEET_TABLE = 'z_CM_Site_Segmentation_Sheet'
CM_SITE_SEGMENTATION_TABLE = 'z_CM_Site_Segmentation'
CM_BROWSER_REPORT_DIRTY_TABLE = 'z_CM_Browser_Report_Dirty'
CM_FLOODLIGHT_TABLE = 'z_CM_Floodlight'


@project.from_parameters
def itp_audit():
  if project.verbose:
    print('ITP Audit Run Queries')

  # Run DV360 related queries
  run_dv_360_queries(project)

  # Create CM Site Segmentation Table and Sheet
  create_cm_site_segmentation(project)

  # CM Segmentation Query
  if project.verbose:
    print('RUN CM Segmentation Query')
  run_query_from_file(Queries.cm_segmentation, CM_BROWSER_REPORT_CLEAN_TABLE)

  run_cm_queries(project)


def run_cm_queries(project):
  # CM Floodlight Lookup Table
  run_query_from_file(Queries.cm_floodlight_join, CM_FLOODLIGHT_TABLE)

  # Floodlight Multichart
  if project.verbose:
    print('RUN CM Floodlight Multichart')
  run_query_from_file(Queries.cm_floodlight_multichart, CM_FLOODLIGHT_MULTICHART_TABLE)


def run_dv_360_queries(project):
  # Create empty DV360 Custom Segments table for join until sheet is created
  table_create('service', project.id, project.task['dataset'],
               DV360_CUSTOM_SEGMENTS_TABLE)

  # Create DV360 Segments Table
  create_dv360_segments(project)

  # Clean DV360 Browser Report
  run_query_from_file(Queries.clean_browser_report, CLEAN_BROWSER_REPORT_TABLE)

  # Browser Performance 2 years
  if project.verbose:
    print('RUN Browser Performance 2 years Query')
  run_query_from_file(Queries.browser_2_year, BROWSER_PERFORMANCE_2YEARS_TABLE)

  #Safari Distribution 90 days
  if project.verbose:
    print('RUN Safari Distribution 90 days Query')
  run_query_from_file(Queries.safari_distribution_90days, SAFARI_DISTRIBUTION_90DAYS_TABLE)

  # Browser Shares Multichart
  if project.verbose:
    print('RUN Dv360 Browser Share Multichart')
  run_query_from_file(Queries.browser_share_multichart, DV360_BROWSER_SHARES_MULTICHART_TABLE)


def create_dv360_segments(project):
  a1_notation = 'A:M'
  schema = [{
      'type': 'STRING',
      'name': 'Advertiser',
      'mode': 'NULLABLE'
  }, {
      'type': 'INTEGER',
      'name': 'Advertiser_Id',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Campaign',
      'mode': 'NULLABLE'
  }, {
      'type': 'INTEGER',
      'name': 'Campaign_Id',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Insertion_Order',
      'mode': 'NULLABLE'
  }, {
      'type': 'INTEGER',
      'name': 'Insertion_Order_Id',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Line_Item',
      'mode': 'NULLABLE'
  }, {
      'type': 'INTEGER',
      'name': 'Line_Item_Id',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Line_Item_Type',
      'mode': 'NULLABLE'
  }, {
      'type': 'INTEGER',
      'name': 'Impressions',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Segment1',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Segment2',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Segment3',
      'mode': 'NULLABLE'
  }]

  sheet_rows = sheets_read(
      'user', project.task['sheet'], 'DV3 Segments', a1_notation, retries=10)

  if not sheet_rows:
    sheet_rows = []

  print('DV360 SEGMENT SHEET TABLE WRITE')

  rows_to_table(
      auth='service',
      project_id=project.id,
      dataset_id=project.task['dataset'],
      table_id=DV360_CUSTOM_SEGMENTS_SHEET_TABLE,
      rows=sheet_rows,
      schema=schema,
      skip_rows=1,
      disposition='WRITE_TRUNCATE')

  # Run Query
  if project.verbose:
    print('RUN DV360 Custom Segments Query')
  run_query_from_file(Queries.dv360_custom_segments, DV360_CUSTOM_SEGMENTS_TABLE)

  # Move Table back to sheets
  query = 'SELECT * from `' + project.id + '.' + project.task[
      'dataset'] + '.' + DV360_CUSTOM_SEGMENTS_TABLE + '`'
  rows = query_to_rows(
      'service', project.id, project.task['dataset'], query, legacy=False)

  # makes sure types are correct in sheet
  a1_notation = a1_notation[:1] + '2' + a1_notation[1:]
  rows = rows_to_type(rows)
  sheets_clear('user', project.task['sheet'], 'DV3 Segments', a1_notation)
  sheets_write('user', project.task['sheet'], 'DV3 Segments', a1_notation, rows)


"""This method will update the BQ table and update the Google sheet
"""


def create_cm_site_segmentation(project):
  # Read sheet to bq table
  sheet_rows = sheets_read(
      'user', project.task['sheet'], 'CM_Site_Segments', 'A:C', retries=10)
  if not sheet_rows:
    sheet_rows = []

  schema = [{
      'type': 'STRING',
      'name': 'Site_Dcm',
      'mode': 'NULLABLE'
  }, {
      'type': 'INTEGER',
      'name': 'Impressions',
      'mode': 'NULLABLE'
  }, {
      'type': 'STRING',
      'name': 'Site_Type',
      'mode': 'NULLABLE'
  }]

  rows_to_table(
      auth='service',
      project_id=project.id,
      dataset_id=project.task['dataset'],
      table_id=CM_SITE_SEGMENTATION_SHEET_TABLE,
      rows=sheet_rows,
      schema=schema,
      skip_rows=1,
      disposition='WRITE_TRUNCATE')

  # Get Site_Type from the sheet
  run_query_from_file(Queries.cm_site_segmentation, CM_SITE_SEGMENTATION_TABLE)

  # Move Table back to sheets
  query = 'SELECT * from `' + project.id + '.' + project.task[
      'dataset'] + '.' + CM_SITE_SEGMENTATION_TABLE + '`'
  rows = query_to_rows(
      'service', project.id, project.task['dataset'], query, legacy=False)

  # makes sure types are correct in sheet
  rows = rows_to_type(rows)
  sheets_clear('user', project.task['sheet'], 'CM_Site_Segments', 'A2:C')
  sheets_write('user', project.task['sheet'], 'CM_Site_Segments', 'A2:C', rows)


def run_query_from_file(query, table_name):
  query_to_table(
    'service',
    project.id,
    project.task['dataset'],
    table_name,
    query.format(**project.task),
    legacy=False
  )


if __name__ == '__main__':
  itp_audit()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

DV360 Audience Analysis

The Audience Wizard Dashboard helps you to track the audience performance across all audiences on Display.

  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->DV360_Audience_Analysis</b> to be created.
  - Join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a> to access the following assets
  - Copy <a href='https://datastudio.google.com/open/1d2vlf4C1roN95NsdsvWNZqKFcYN8N9Jg' target='_blank'>Sample DV360 Audience Analysis Dataset</a>.
  - Click Edit Connection, and change to <b>BigQuery->UNDEFINED->UNDEFINED->DV360_Audience_Analysis</b>.
  - Copy <a href='https://datastudio.google.com/open/1Ij_RluqolElm7Nny9fBrIAPRB9ObUl0M' target='_blank'>Sample DV360 Audience Analysis Report</a>.
  - When prompted choose the new data source you just created.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_slug': '',  # Place where tables will be created in BigQuery.
  'recipe_timezone': 'America/Los_Angeles',  # Timezone for report dates.
  'recipe_name': '',  # Name of report in DV360, should be unique.
  'partners': [],  # DV360 partner id.
  'advertisers': [],  # Comma delimited list of DV360 advertiser ids.
  'recipe_project': '',  # Google Cloud Project Id.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      2,
      4,
      6
    ]
  },
  'tasks': [
    {
      'dataset': {
        'hour': [
          1
        ],
        'auth': 'service',
        'description': 'Create a dataset for bigquery tables.',
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'dataRange': 'LAST_7_DAYS',
              'format': 'CSV',
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Audience Analysis Performance ',
                  'description': 'Name of report in DV360, should be unique.'
                }
              }
            },
            'params': {
              'type': 'TYPE_GENERAL',
              'groupBys': [
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER',
                'FILTER_AUDIENCE_LIST',
                'FILTER_USER_LIST',
                'FILTER_AUDIENCE_LIST_TYPE',
                'FILTER_AUDIENCE_LIST_COST',
                'FILTER_PARTNER_CURRENCY'
              ],
              'metrics': [
                'METRIC_IMPRESSIONS',
                'METRIC_CLICKS',
                'METRIC_TOTAL_CONVERSIONS',
                'METRIC_LAST_CLICKS',
                'METRIC_LAST_IMPRESSIONS',
                'METRIC_TOTAL_MEDIA_COST_PARTNER'
              ]
            }
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          6
        ],
        'auth': 'user',
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Audience Analysis Performance ',
              'description': 'Name of report in DV360, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'DV360_Audience_Performance',
            'header': True,
            'schema': [
              {
                'mode': 'REQUIRED',
                'name': 'advertiser',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'advertiser_id',
                'type': 'INT64'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list_id',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_type',
                'type': 'STRING'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_cost_usd',
                'type': 'STRING'
              },
              {
                'mode': 'NULLABLE',
                'name': 'partner_currency',
                'type': 'STRING'
              },
              {
                'mode': 'NULLABLE',
                'name': 'impressions',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'clicks',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'total_conversions',
                'type': 'FLOAT'
              },
              {
                'mode': 'NULLABLE',
                'name': 'post_click_conversions',
                'type': 'FLOAT'
              },
              {
                'mode': 'NULLABLE',
                'name': 'post_view_conversions',
                'type': 'FLOAT'
              },
              {
                'mode': 'NULLABLE',
                'name': 'total_media_cost_partner_currency',
                'type': 'FLOAT'
              }
            ]
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'dataRange': 'LAST_7_DAYS',
              'format': 'CSV',
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Audience Analysis First Party',
                  'description': 'Name of report in DV360, should be unique.'
                }
              }
            },
            'params': {
              'type': 'TYPE_INVENTORY_AVAILABILITY',
              'groupBys': [
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER',
                'FILTER_USER_LIST_FIRST_PARTY_NAME',
                'FILTER_USER_LIST_FIRST_PARTY',
                'FILTER_FIRST_PARTY_AUDIENCE_LIST_TYPE',
                'FILTER_FIRST_PARTY_AUDIENCE_LIST_COST'
              ],
              'metrics': [
                'METRIC_BID_REQUESTS',
                'METRIC_UNIQUE_VISITORS_COOKIES'
              ]
            }
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          6
        ],
        'auth': 'user',
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Audience Analysis First Party',
              'description': 'Name of report in DV360, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'DV360_First_Party_Audience',
            'header': True,
            'schema': [
              {
                'mode': 'REQUIRED',
                'name': 'advertiser',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'advertiser_id',
                'type': 'INT64'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list_id',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_type',
                'type': 'STRING'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_cost_usd',
                'type': 'FLOAT'
              },
              {
                'mode': 'NULLABLE',
                'name': 'potential_impressions',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'unique_cookies_with_impressions',
                'type': 'INT64'
              }
            ]
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'dataRange': 'LAST_7_DAYS',
              'format': 'CSV',
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Audience Analysis Google',
                  'description': 'Name of report in DV360, should be unique.'
                }
              }
            },
            'params': {
              'type': 'TYPE_INVENTORY_AVAILABILITY',
              'groupBys': [
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER',
                'FILTER_AUDIENCE_LIST',
                'FILTER_USER_LIST',
                'FILTER_AUDIENCE_LIST_TYPE',
                'FILTER_AUDIENCE_LIST_COST'
              ],
              'metrics': [
                'METRIC_BID_REQUESTS',
                'METRIC_UNIQUE_VISITORS_COOKIES'
              ]
            }
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          6
        ],
        'auth': 'user',
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Audience Analysis Google',
              'description': 'Name of report in DV360, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'DV360_Google_Audience',
            'header': True,
            'schema': [
              {
                'mode': 'REQUIRED',
                'name': 'advertiser',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'advertiser_id',
                'type': 'INT64'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list_id',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_type',
                'type': 'STRING'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_cost_usd',
                'type': 'FLOAT'
              },
              {
                'mode': 'NULLABLE',
                'name': 'potential_impressions',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'unique_cookies_with_impressions',
                'type': 'INT64'
              }
            ]
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'dataRange': 'LAST_7_DAYS',
              'format': 'CSV',
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Audience Analysis Third Party',
                  'description': 'Name of report in DV360, should be unique.'
                }
              }
            },
            'params': {
              'type': 'TYPE_INVENTORY_AVAILABILITY',
              'groupBys': [
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER',
                'FILTER_USER_LIST_THIRD_PARTY_NAME',
                'FILTER_USER_LIST_THIRD_PARTY',
                'FILTER_THIRD_PARTY_AUDIENCE_LIST_TYPE',
                'FILTER_THIRD_PARTY_AUDIENCE_LIST_COST'
              ],
              'metrics': [
                'METRIC_BID_REQUESTS',
                'METRIC_UNIQUE_VISITORS_COOKIES'
              ]
            }
          }
        }
      }
    },
    {
      'dbm': {
        'hour': [
          6
        ],
        'auth': 'user',
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Audience Analysis Third Party',
              'description': 'Name of report in DV360, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'DV360_Third_Party_Audience',
            'header': True,
            'schema': [
              {
                'mode': 'REQUIRED',
                'name': 'advertiser',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'advertiser_id',
                'type': 'INT64'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list',
                'type': 'STRING'
              },
              {
                'mode': 'REQUIRED',
                'name': 'audience_list_id',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_type',
                'type': 'STRING'
              },
              {
                'mode': 'NULLABLE',
                'name': 'audience_list_cost_usd',
                'type': 'FLOAT'
              },
              {
                'mode': 'NULLABLE',
                'name': 'potential_impressions',
                'type': 'INT64'
              },
              {
                'mode': 'NULLABLE',
                'name': 'unique_cookies_with_impressions',
                'type': 'INT64'
              }
            ]
          }
        }
      }
    },
    {
      'bigquery': {
        'hour': [
          6
        ],
        'auth': 'service',
        'from': {
          'query': " SELECT   p.advertiser_id,   p.advertiser,   p.audience_list_id,   IF (p.audience_list_type = 'Bid Manager Audiences', 'Google', p.audience_list_type) AS audience_list_type,   CASE     WHEN REGEXP_CONTAINS(p.audience_list, 'Affinity') THEN 'Affinity'     WHEN REGEXP_CONTAINS(p.audience_list, 'Demographics') THEN 'Demographics'     WHEN REGEXP_CONTAINS(p.audience_list, 'In-Market') THEN 'In-Market'     WHEN REGEXP_CONTAINS(p.audience_list, 'Similar') THEN 'Similar'     ELSE 'Custom'   END AS audience_list_category,   p.audience_list,   IF(p.audience_list_cost_usd = 'Unknown', 0.0, CAST(p.audience_list_cost_usd AS FLOAT64)) AS audience_list_cost,   p.total_media_cost_partner_currency AS total_media_cost,   p.impressions,   p.clicks,   p.total_conversions,   COALESCE(ggl.potential_impressions, fst.potential_impressions, trd.potential_impressions) AS potential_impressions,   COALESCE(ggl.unique_cookies_with_impressions, fst.unique_cookies_with_impressions, trd.unique_cookies_with_impressions) AS potential_reach FROM   `[PARAMETER].[PARAMETER].DV360_Audience_Performance` p LEFT JOIN   `[PARAMETER].[PARAMETER].DV360_Google_Audience` ggl   USING (advertiser_id, audience_list_id) LEFT JOIN   `[PARAMETER].[PARAMETER].DV360_First_Party_Audience` fst   USING (advertiser_id, audience_list_id) LEFT JOIN   `[PARAMETER].[PARAMETER].DV360_Third_Party_Audience` trd   USING (advertiser_id, audience_list_id) ",
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'order': 6,
                'default': '',
                'description': 'Google Cloud Project Id.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'order': 6,
                'default': '',
                'description': 'Google Cloud Project Id.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'order': 6,
                'default': '',
                'description': 'Google Cloud Project Id.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'order': 6,
                'default': '',
                'description': 'Google Cloud Project Id.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            }
          ],
          'legacy': False
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be created in BigQuery.'
            }
          },
          'view': 'DV360_Audience_Analysis'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('audience_analysis', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
# 
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery

class DV360_Targeting:

  def __init__(self):
    self.schema = Discovery_To_BigQuery(
      'displayvideo',
      'v1'
    )

  def get_schema(self, name, resource):
    return [
      { 'name': 'partnerId', 'type': 'INTEGER', 'mode': 'NULLABLE' },
      { 'name': 'advertiserId', 'type': 'INTEGER', 'mode': 'NULLABLE' },
      { 'name': 'lineItemId', 'type': 'INTEGER', 'mode': 'NULLABLE' },
      { 'name': 'name', 'type': 'STRING', 'mode': 'REQUIRED' },
      { 'name': 'assignedTargetingOptionId', 'type': 'STRING', 'mode': 'REQUIRED' },
      { 'name': 'targetingType', 'type': 'STRING', 'mode': 'REQUIRED' },
      { 'name': 'inheritance', 'type': 'STRING', 'mode': 'REQUIRED' },
      { 'name': name, 'type':'RECORD', 'mode':'REQUIRED', 'fields':self.schema.resource_schema(resource) }
    ]

  def get_partner(self, auth, partners, targeting_type):
    for partner in partners:
      for row in API_DV360(
        auth,
        iterate=True
      ).partners().targetingTypes().assignedTargetingOptions().list(
        partnerId=str(partner),
        targetingType=targeting_type
      ).execute():
        row['partnerId'] = partner
        row['advertiserId'] = None
        row['lineItemId'] = None
        yield row


  def get_advertiser(self, auth, advertisers, targeting_type):
    for advertiser in advertisers:
      for row in API_DV360(
        auth,
        iterate=True
      ).advertisers().targetingTypes().assignedTargetingOptions().list(
        advertiserId=str(advertiser),
        targetingType=targeting_type
      ).execute():
        row['partnerId'] = None
        row['advertiserId'] = advertiser
        row['lineItemId'] = None
        yield row


  def get_line_item(self, auth, parameters, targeting_type):
    for parameter in parameters:
      for row in API_DV360(
        auth,
        iterate=True
      ).advertisers().lineItems().targetingTypes().assignedTargetingOptions().list(
        advertiserId=str(parameter['advertiserId']),
        lineItemId=str(parameter['lineItemId']),
        targetingType=targeting_type
      ).execute():
        row['partnerId'] = None
        row['advertiserId'] = parameter['lineItemId']
        row['lineItemId'] = None
        print(row)
        yield row


TARGETING_TYPES = {
  'TARGETING_TYPE_CHANNEL':{
    'name':'channelDetails',
    'resource':'ChannelAssignedTargetingOptionDetails',
    'partner':True,
    'advertiser':True,
    'lineitem':True,
  },
  'TARGETING_TYPE_APP_CATEGORY':{
    'name':'appCategoryDetails',
    'resource':'AppCategoryAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_APP':{
    'name':'appDetails',
    'resource':'AppAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_URL':{
    'name':'urlDetails',
    'resource':'UrlAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_DAY_AND_TIME':{
    'name':'dayAndTimeDetails',
    'resource':'DayAndTimeAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_AGE_RANGE':{
    'name':'ageRangeDetails',
    'resource':'AgeRangeAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_REGIONAL_LOCATION_LIST':{
    'name':'regionalLocationListDetails',
    'resource':'RegionalLocationListAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_PROXIMITY_LOCATION_LIST':{
    'name':'proximityLocationListDetails',
    'resource':'ProximityLocationListAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_GENDER':{
    'name':'genderDetails',
    'resource':'GenderAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_VIDEO_PLAYER_SIZE':{
    'name':'videoPlayerSizeDetails',
    'resource':'VideoPlayerSizeAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_USER_REWARDED_CONTENT':{
    'name':'userRewardedContentDetails',
    'resource':'UserRewardedContentAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_PARENTAL_STATUS':{
    'name':'parentalStatusDetails',
    'resource':'ParentalStatusAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_CONTENT_INSTREAM_POSITION':{
    'name':'contentInstreamPositionDetails',
    'resource':'ContentInstreamPositionAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_CONTENT_OUTSTREAM_POSITION':{
    'name':'contentOutstreamPositionDetails',
    'resource':'ContentOutstreamPositionAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_DEVICE_TYPE':{
    'name':'deviceTypeDetails',
    'resource':'DeviceTypeAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_AUDIENCE_GROUP':{
    'name':'audienceGroupDetails',
    'resource':'AudienceGroupAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_BROWSER':{
    'name':'browserDetails',
    'resource':'BrowserAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_HOUSEHOLD_INCOME':{
    'name':'householdIncomeDetails',
    'resource':'HouseholdIncomeAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_ON_SCREEN_POSITION':{
    'name':'onScreenPositionDetails',
    'resource':'OnScreenPositionAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_CARRIER_AND_ISP':{
    'name':'carrierAndIspDetails',
    'resource':'CarrierAndIspAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_KEYWORD':{
    'name':'keywordDetails',
    'resource':'KeywordAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_NEGATIVE_KEYWORD_LIST':{
    'name':'negativeKeywordListDetails',
    'resource':'NegativeKeywordListAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_OPERATING_SYSTEM':{
    'name':'operatingSystemDetails',
    'resource':'OperatingSystemAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_DEVICE_MAKE_MODEL':{
    'name':'deviceMakeModelDetails',
    'resource':'DeviceMakeModelAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_ENVIRONMENT':{
    'name':'environmentDetails',
    'resource':'EnvironmentAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_INVENTORY_SOURCE':{
    'name':'inventorySourceDetails',
    'resource':'InventorySourceAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_CATEGORY':{
    'name':'categoryDetails',
    'resource':'CategoryAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_VIEWABILITY':{
    'name':'viewabilityDetails',
    'resource':'ViewabilityAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_AUTHORIZED_SELLER_STATUS':{
    'name':'authorizedSellerStatusDetails',
    'resource':'AuthorizedSellerStatusAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_LANGUAGE':{
    'name':'languageDetails',
    'resource':'LanguageAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_GEO_REGION':{
    'name':'geoRegionDetails',
    'resource':'GeoRegionAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_INVENTORY_SOURCE_GROUP':{
    'name':'inventorySourceGroupDetails',
    'resource':'InventorySourceGroupAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_DIGITAL_CONTENT_LABEL_EXCLUSION':{
    'name':'digitalContentLabelExclusionDetails',
    'resource':'DigitalContentLabelAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':True,
    'lineitem':True,
  },
  'TARGETING_TYPE_SENSITIVE_CATEGORY_EXCLUSION':{
    'name':'sensitiveCategoryExclusionDetails',
    'resource':'SensitiveCategoryAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':True,
    'lineitem':True,
  },
  'TARGETING_TYPE_EXCHANGE':{
    'name':'exchangeDetails',
    'resource':'ExchangeAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_SUB_EXCHANGE':{
    'name':'subExchangeDetails',
    'resource':'SubExchangeAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
  'TARGETING_TYPE_THIRD_PARTY_VERIFIER':{
    'name':'thirdPartyVerifierDetails',
    'resource':'ThirdPartyVerifierAssignedTargetingOptionDetails',
    'partner':False,
    'advertiser':False,
    'lineitem':True,
  },
}



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# https://developers.google.com/drive/api/v3/manage-uploads
# https://developers.google.com/drive/api/v3/reference/about#methods

import re
import mimetypes
from io import BytesIO
from googleapiclient.http import MediaIoBaseUpload
from googleapiclient.errors import HttpError

from starthinker.config import BUFFER_SCALE
from starthinker.util.project import project
from starthinker.util.google_api import API_Retry, API_Drive

CHUNKSIZE = int(200 * 1024000 *
                BUFFER_SCALE)  # scale is controlled in config.py

#https://drive.google.com/open?id=1um2RlasnKTvF_I6uVUsbmk4ioyW1Zemj
#https://drive.google.com/open?id=1NQH5-yP9LM2dGTuGvdOA31KO1OwblruC-dK92F5IVLE
#https://drive.google.com/open?id=1IM-Dnq5Pc1Lje6eTKrN1ubpclRcbvAROrmSn23xl7UI
#https://drive.google.com/open?id=1XMFEUBxpRuAy8qG5tg3AM0NVm9bklpCc6SelZHpPj-g

#https://docs.google.com/document/d/1XMFEUBxpRuAy8qG5tg3AM0NVm9bklpCc6SelZHpPj-g/edit
#https://docs.google.com/presentation/d/1IM-Dnq5Pc1Lje6eTKrN1ubpclRcbvAROrmSn23xl7UI/edit
#https://docs.google.com/spreadsheets/d/1YU8zXzmeqgQhnnRIs4FpNKfbZ2B_aN77XbSKHBEa_gc/edit
#https://datastudio.google.com/c/reporting/0B1dzmNeF4E-MZ21Zb1h6SnRoc28/page/L1MF
#https://datastudio.google.com/c/datasources/1a6K-XdPUzCYRXZp1ZcmeOUOURc9wn2Jj


def about(auth, fields='importFormats'):
  response = API_Drive(auth).about().get(fields=fields).execute()
  return response


def file_id(auth, url_or_name):

  if url_or_name.startswith('https://drive.google.com/open?id='):
    return url_or_name.split('?id=', 1)[-1]

  elif url_or_name.startswith('https://docs.google.com/'):
    m = re.search(
        '^(?:https:\/\/docs.google.com\/\w+\/d\/)([a-zA-Z0-9-_]+)(?:\/.*)?$',
        url_or_name)
    if m:
      return m.group(1)

  elif url_or_name.startswith('https://datastudio.google.com/'):
    m = re.search(
        '^(?:https:\/\/datastudio.google.com\/c\/\w+\/)([a-zA-Z0-9-_]+)(?:\/.*)?$',
        url_or_name)
    if m:
      return m.group(1)

  # check if name given convert to ID "Some Document"
  else:
    document = file_find(auth, url_or_name)
    if document:
      return document['id']

      # check if just ID given, "1uN9tnb-DZ9zZflZsoW4_34sf34tw3ff"
    else:
      m = re.search('^([a-zA-Z0-9-_]+)$', url_or_name)
      if m:
        return m.group(1)

  # probably a mangled id or name does not exist
  if project.verbose:
    print('DOCUMENT DOES NOT EXIST', url_or_name)
  return None


def file_get(auth, drive_id):
  return API_Drive(auth).files().get(fileId=drive_id).execute()


def file_exists(auth, name):
  drive_id = file_id(auth, name)
  if drive_id:
    try:
      API_Drive(auth).files().get(fileId=drive_id).execute()
      return True
    except HttpError:
      return False
  return False


def file_find(auth, name, parent=None):
  query = "trashed = false and name = '%s'" % name
  if parent:
    query = "%s and '%s' in parents" % (query, parent)

  try:
    return next(API_Drive(auth, iterate=True).files().list(q=query).execute())
  except StopIteration:
    return None


def file_delete(auth, name, parent=None):
  drive_id = file_id(auth, name)

  if drive_id:
    API_Drive(auth).files().delete(fileId=drive_id).execute()
    return True
  else:
    return False


def file_create(auth, name, filename, data, parent=None):
  """ Checks if file with name already exists ( outside of trash ) and

    if not, uploads the file.  Determines filetype based on filename extension
    and attempts to map to Google native such as Docs, Sheets, Slides, etc...

    For example:
    -  ```file_create('user', 'Sample Document', 'sample.txt', BytesIO('File
    contents'))```
    -  Creates a Google Document object in the user's drive.

    -  ```file_Create('user', 'Sample Sheet', 'sample.csv',
    BytesIO('col1,col2\nrow1a,row1b\n'))````
    -  Creates a Google Sheet object in the user's drive.

    See: https://developers.google.com/drive/api/v3/manage-uploads

    ### Args:
    -  * auth: (string) specify 'service' or 'user' to toggle between
    credentials used to access
    -  * name: (string) name of file to create, used as key to check if file
    exists
    -  * filename: ( string) specified as "file.extension" only to automate
    detection of mime type.
    -  * data: (BytesIO) any file like object that can be read from
    -  * parent: (string) the Google Drive to upload the file to

    ### Returns:
    -  * JSON specification of file created or existing.

    """

  # attempt to find the file by name ( not in trash )
  drive_file = file_find(auth, name, parent)

  # if file exists, return it, prevents obliterating user changes
  if drive_file:
    if project.verbose:
      print('Drive: File exists.')

  # if file does not exist, create it
  else:
    if project.verbose:
      print('Drive: Creating file.')

    # file mime is used for uplaod / fallback
    # drive mime attempts to map to a native Google format
    file_mime = mimetypes.guess_type(filename, strict=False)[0]
    drive_mime = about('importFormats')['importFormats'].get(
        file_mime, file_mime)[0]

    if project.verbose:
      print('Drive Mimes:', file_mime, drive_mime)

    # construct upload object, and stream upload in chunks
    body = {
        'name': name,
        'parents': [parent] if parent else [],
        'mimeType': drive_mime,
    }

    media = MediaIoBaseUpload(
        BytesIO(data or ' '),  # if data is empty BAD REQUEST error occurs
        mimetype=file_mime,
        chunksize=CHUNKSIZE,
        resumable=True)

    drive_file = API_Drive(auth).files().create(
        body=body, media_body=media, fields='id').execute()

  return drive_file


def file_copy(auth, source_name, destination_name):
  destination_id = file_id(auth, destination_name)

  if destination_id:
    if project.verbose:
      print('Drive: File exists.')
    return file_get(auth, destination_id)

  else:
    source_id = file_id(auth, source_name)

    if source_id:
      body = {'visibility': 'PRIVATE', 'name': destination_name}
      return API_Drive(auth).files().copy(fileId=source_id, body=body).execute()
    else:
      return None


def folder_create(auth, name, parent=None):
  body = {
      'name': name,
      'parents': [parent] if parent else [],
      'mimeType': 'application/vnd.google-apps.folder'
  }
  return API_Drive(auth).files().create(body=body, fields='id').execute()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
from datetime import datetime

MARKUP_TO_HTML = re.compile(r'\[(.*?)\]\((.*?)\)', flags=re.IGNORECASE)
MARKUP_TO_TEXT = re.compile(r'\[(.*?)\]\((.*?)\)', flags=re.IGNORECASE)


class EmailTemplate:

  def __init__(self, template={}):
    self.content_html = ''
    self.content_text = ''

    self.style = {
        'background':
            template.get('style', {}).get('background', '#f2f2f2'),
        'foreground':
            template.get('style', {}).get('foreground', '#ffffff'),
        'text':
            template.get('style', {}).get('text', '#414347'),
        'link':
            template.get('style', {}).get('link', '#4285f4'),
        'font':
            template.get('style',
                         {}).get('font',
                                 'Roboto, Helvetica, Arial sans-serif;'),
        'align':
            template.get('style', {}).get('align', 'left'),
    }

    self.subject = template.get('subject', '')
    self.logo = template.get('logo', '')
    self.footer = template.get('footer', [])
    self.copyright = template.get('copyright',
                                  'Copyright &copy; %d' % datetime.today().year)

    if template:
      self.template(template)

  def template(self, template):

    if 'body' in template:
      self.template(template['body'])
    if 'greeting' in template:
      self.greeting(**template['greeting'])
    if 'header' in template:
      self.header(template['header'])
    if 'image' in template:
      self.image(**template['image'])
    if 'paragraph' in template:
      self.paragraph(template['paragraph'])
    if 'button' in template:
      self.button(**template['button'])
    if 'table' in template:
      self.table(**template['table'])
    if 'list' in template:
      self.list(template['list'])
    if 'sections' in template:
      for s in template['sections']:
        self.section(True, background=s.get('background', None))
        self.template(s)
        self.section()
    if 'grid' in template:
      self.grid(True)
      for r in template['grid']:
        self.row(True)
        for c in r:
          self.column(len(r))
          self.template(c)
          self.column()
        self.row()
      self.grid()

  def _header_css(self):
    return ('width:100%%;margin:10px 0px 20px '
            '0px;padding:0px;border:0px;text-align:%(align)s;color:%(text)s;font-family:%(font)s;font-size:22px;line-height:30px;font-weight:500;letter-spacing:-0.31px;border:0px;') % self.style

  def _text_css(self):
    return ('width:100%%;margin:10px '
            '0px;padding:0px;border:0px;font-weight:normal;text-align:%(align)s;color:%(text)s;font-family:%(font)s;font-size:15px;line-height:26px;') % self.style

  def _link_css(self):
    return 'color:%(link)s;text-decoration:none;' % self.style

  def _button_css(self):
    return ('display:inline-block;margin:10px '
            'auto;border-radius:3px;-webkit-border-radius:3px;-moz-border-radius:3px;padding:11px'
            ' '
            '19px;background-color:%(link)s;color:%(background)s;font-family:%(font)s;font-size:16px;font-weight:bold;letter-spacing:-0.31px;text-decoration:none;text-align:center;') % self.style

  def _table_css(self):
    return 'color:%(text)s;font-family:%(font)s;font-size:12px;line-height:16px;' % self.style

  def markup_to_html(self, markup):
    return MARKUP_TO_HTML.sub(
        r'<a href="\2" style="%s">\1</a>' % self._link_css(),
        markup).replace('\n', '<br/><br/>')

  def markup_to_text(self, markup):
    return MARKUP_TO_TEXT.sub(r'\1 (\2)', markup).replace('\n', '<br/><br/>')

  def section(self, on=False, background=None):
    value = {
        'image':
            'background-image:url(\'%s\');background-repeat:no-repeat;background-size:contain;'
            % background if background else '',
        'align':
            self.style['align']
    }

    # HTML
    if on:
      self.content_html += ('<div style="margin:0px '
                            '15px;padding:1px;text-align:%(align)s;%(image)s">') % value
    else:
      self.content_html += '</div>'
      self.content_text += '\n'

  def align(self, text_align):
    self.style['align'] = text_align

  def greeting(self, name='', salutation='Hi'):
    if name:
      # HTML
      self.content_html += '<p style="%s">%s %s,<p>' % (self._text_css(),
                                                        salutation, name)

      # Text
      self.content_text += '%s %s,\n\n' % (salutation, name)

    else:
      # HTML
      self.content_html += '<p style="%s">%s,<p>' % (self._text_css(),
                                                     salutation)

      # Text
      self.content_text += '%s,\n\n' % salutation

  def header(self, text):
    # HTML
    self.content_html += '<p style="%s">%s</p>' % (self._header_css(),
                                                   self.markup_to_html(text))

    # Text
    self.content_text += '\n--------------------------------------------------\n\n%s\n\n' % self.markup_to_text(
        text)

  def paragraph(self, text):

    # HTML
    self.content_html += '<p style="%s">%s<p>' % (
        self._text_css(), self.markup_to_html(text)
        #MARKUP_TO_HTML.sub(r'<a href="\2" style="%s">\1</a>' % self._link_css(), text).replace('\n', '<br/><br/>')
    )

    # Text
    #self.content_text += '%s\n\n' % MARKUP_TO_TEXT.sub(r'\1 (\2)', text).replace('\n', '<br/><br/>')
    self.content_text += '%s\n\n' % self.markup_to_text(text)

  def image(self, src, link=None):
    # HTML
    if link:
      self.content_html += '<a href="%s">' % link
    self.content_html += ('<img src="%s" '
                          'style="max-width:100%%;width:auto;height:auto;margin:10px'
                          ' auto;padding:0px;border:0px;"/>') % src
    if link:
      self.content_html += '</a>'

    # Text
    self.content_text += 'IMAGE: %s\n' % src
    if link:
      self.content_text += 'LINK: %s\n\n' % link

  def button(self, text, link, big=False):
    # HTML
    self.content_html += ('<p style="width:100%%;margin:0px '
                          '0px;padding:0px;border:0px;"><a href="%s" '
                          'style="%s">%s</a></p>') % (
        link, self._button_css(), text)

    # Text
    self.content_text += '%s: %s' % (text, link)

  def grid(self, on=False):
    if on:
      self.content_html += ('<table style="width:100%;margin:20px '
                            '0px;padding:0px;border:0px;border-spacing:0px '
                            '20px;">')
    else:
      self.content_html += '</table>'

  def row(self, on=False):
    if on:
      self.content_html += '<tr>'
    else:
      self.content_html += '</tr>'

  def column(self, number=0):
    if number:
      width = int(100 / number)
      self.content_html += ('<td '
                            'style="width:%d%%;padding:10px;text-align:center;vertical-align:top;">') % width
    else:
      self.content_html += '</td>'

  def display(self, header, paragraph, image, link, button):
    # HTML
    self.grid(True)
    self.row(True)

    if image:
      self.column(True)
      self.image(image, link)
      self.column()

    self.column(True)
    if header:
      self.header(header)
    if paragraph:
      self.paragraph(paragraph)
    if button:
      self.button(button, link)
    self.column()

    self.row()
    self.grid()

  def list(self, elements):
    # HTML
    border = 0
    self.content_html += ('<table style="margin:20px auto 0px '
                          'auto;padding:0px;border:0px;border-collapse:collapse;">')
    for element in elements:
      self.content_html += ('<tr><td '
                            'style="padding:10px;text-align:%s;border-top:%dpx '
                            'solid #ccc;%s">%s</td></tr>') % (
          self.style['align'], border, self._table_css(),
          self.markup_to_html(element))
      border = 1
    self.content_html += '</table>'

    # Text
    self.content_text += '\n\n'
    for element in elements:
      self.content_text += '  - %s\n' % self.markup_to_text(element)
    self.content_text == '\n'

  def table(self, schema, rows):
    aligns = {
        'BOOLEAN': 'center',
        'INTEGER': 'right',
        'FLOAT': 'right',
        'STRING': 'left',
        'URL': 'center'
    }

    # HTML
    self.content_html += ('<table style="width:100%;margin:20px '
                          '0px;padding:0px;border:0px;border-collapse:collapse;">')
    self.content_html += '<tr>'

    # headers
    for header in schema:
      self.content_html += ('<td '
                            'style="padding:10px;text-align:%s;border-bottom:1px'
                            ' solid #ccc;font-weight:bold;%s">%s</td>') % (
          aligns.get(header['type'], 'left'), self._table_css(), header['name'])
    self.content_html += '</tr>'

    # rows
    for row in rows:
      self.content_html += '<tr>'
      for column, header in enumerate(schema):
        value = row[column]
        # change links into clickable buttons ( do we want to make up our own schema here? for images too? )
        if isinstance(value, str) and value.startswith('http'):
          value = ('<a href="%s" style="border-radius:50px;padding:1px '
                   '12px;%s">&rarr;</a>') % (
              value, self._button_css())
        self.content_html += ('<td '
                              'style="padding:10px;text-align:%s;border-bottom:1px'
                              ' solid #ccc;%s">%s</td>') % (
            aligns.get(header['type'], 'left'), self._table_css(), value)
      self.content_html += '</tr>'

    self.content_html += '</table>'

    # Text
    for row in rows:
      self.content_text += '--------------------------------------------------\n'
      for column, header in enumerate(schema):
        self.content_text += '%s: %s\n' % (header['name'], row[column])
      self.content_text += '\n'

  def get_subject(self):
    return self.subject

  def get_html(self):
    logo = ''
    if self.logo:
      logo += ('<p style="width:100%%;margin:10px auto;padding:0.1em '
               '0px;border:0px;text-align:left;">')
      logo += '<img src="%s" style="margin:10px 0px;padding:0px;border:0px"/>' % self.logo
      logo += '</p>'

    footer = ''
    if self.footer:
      footer += ('<p style="width:100%;margin:30px '
                 'auto;padding:0px;border:0px;text-align:center;">')
      for f in self.footer:
        footer += ('<a href="%s" style="display:inline-block;margin: '
                   '0px;padding:0px '
                   '10px;color:%s;font-family:%s;font-size:15px;font-weight:normal;line-height:32px;text-decoration:none;">%s</a>') % (
            f['link'], self.style['text'], self.style['font'], f['text'])
      footer += '</p>'
    if self.copyright:
      footer += ('<p style="width:100%%;margin:20px '
                 'auto;padding:0px;border:0px;font-weight:bold;text-align:center;color:%(text)s;font-family:%(font)s;font-size:15px;line-height:32px;">') % self.style
      footer += self.copyright
      footer += '</p>'

    return """
      <html>
        <body>
          <div style="width:100%%;padding:40px 0px;margin:0px;background-color:%s;">
            <div style="max-width:600px;width:100%%;padding:0px;margin:0px auto;">
              %s
              <table style="width:100%%;margin:0px auto;text-align:left;background-color:%s">
                <tr><td>%s</td></tr>
              </table>
              %s
            </div>
          </div>
        </body>
      </html>
    """ % (self.style['background'], logo, self.style['foreground'],
           self.content_html, footer)

  def get_text(self):
    footer = ''
    if self.footer:
      footer += '\n--------------------------------------------------\n'
      for f in self.footer:
        footer += '\n%s: %s' % (f['text'], f['link'])
    if self.copyright:
      footer += '\n\n--------------------------------------------------\n\n%s' % self.copyright

    return self.content_text + footer


if __name__ == '__main__':

  t = EmailTemplate()
  t.header('This Is A Header')
  t.paragraph(
      'A paragraph is a few sentences of text.  Although you can use HTML, it will break the text version.'
  )
  t.display('Section Header', 'A section paragraph is a few sentences of text.',
            'http://www.image.com/image.jpg', 'http://www.google.com',
            'A Button')
  t.table([
      {
          'name': 'Column A',
          'type': 'STRING'
      },
      {
          'name': 'Column B',
          'type': 'INTEGER'
      },
      {
          'name': 'Column C',
          'type': 'BOOLEAN'
      },
      {
          'name': 'Column D',
          'type': 'FLOAT'
      },
      {
          'name': 'Column E',
          'type': 'URL'
      },
  ], [
      ['First', 1, True, 1.1, 'http://www.google.com'],
      ['Second', 2, False, 2.2, 'http://www.google.com'],
      ['Third', 3, True, 3.3, 'http://www.google.com'],
  ])
  print(t.get_html())
  print(t.get_text())



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Target_Schema = [
    {
        'name': 'criteria_id',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'parameter',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.insertion_order import insertion_order_commit
from starthinker.task.dv_editor.line_item import line_item_commit
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def pacing_clear():
  sheets_clear(project.task["auth_sheets"], project.task["sheet"], "Pacing", "A2:Z")


def pacing_load():

  # write pacings to sheet
  rows = get_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset":
                  project.task["dataset"],
              "query":
                  """SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         NULL,
         I.pacing.pacingPeriod,
         I.pacing.pacingPeriod,
         I.pacing.pacingType,
         I.pacing.pacingType,
         CAST(I.pacing.dailyMaxMicros AS INT64) / 100000,
         CAST(I.pacing.dailyMaxMicros AS INT64) / 100000,
         I.pacing.dailyMaxImpressions,
         I.pacing.dailyMaxImpressions
       FROM `{dataset}.DV_InsertionOrders` AS I
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON I.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON I.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       UNION ALL
       SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         CONCAT(L.displayName, ' - ', L.lineItemId),
         L.pacing.pacingPeriod,
         L.pacing.pacingPeriod,
         L.pacing.pacingType,
         L.pacing.pacingType,
         CAST(L.pacing.dailyMaxMicros AS INT64) / 100000,
         CAST(L.pacing.dailyMaxMicros AS INT64) / 100000,
         L.pacing.dailyMaxImpressions,
         L.pacing.dailyMaxImpressions
       FROM `{dataset}.DV_LineItems` AS L
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON L.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_InsertionOrders` AS I
       ON L.insertionOrderId=I.insertionOrderId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON L.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       """.format(**project.task),
              "legacy":
                  False
          }
      })

  put_rows(project.task["auth_sheets"], {
      "sheets": {
          "sheet": project.task["sheet"],
          "tab": "Pacing",
          "range": "A2"
      }
  }, rows)


def pacing_audit():
  rows = get_rows(project.task["auth_sheets"], {
      "sheets": {
          "sheet": project.task["sheet"],
          "tab": "Pacing",
          "range": "A2:Z"
      }
  })

  put_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset": project.task["dataset"],
              "table": "SHEET_Pacing",
              "schema": [
                  { "name": "Partner", "type": "STRING" },
                  { "name": "Advertiser", "type": "STRING" },
                  { "name": "Campaign", "type": "STRING" },
                  { "name": "Insertion_Order", "type": "STRING" },
                  { "name": "Line_Item", "type": "STRING" },
                  { "name": "Period", "type": "STRING" },
                  { "name": "Period_Edit", "type": "STRING" },
                  { "name": "Type", "type": "STRING" },
                  { "name": "Type_Edit", "type": "STRING" },
                  { "name": "Daily_Budget", "type": "FLOAT" },
                  { "name": "Daily_Budget_Edit", "type": "FLOAT" },
                  { "name": "Daily_Impressions", "type": "INTEGER" },
                  { "name": "Daily_Impressions_Edit", "type": "INTEGER" },
              ],
              "format": "CSV"
          }
      }, rows)

  query_to_view(
      project.task["auth_bigquery"],
      project.id,
      project.task["dataset"],
      "AUDIT_Pacing",
      """WITH
      /* Check if sheet values are set */
      INPUT_ERRORS AS (
        SELECT
        *
        FROM (
          SELECT
            'Pacing' AS Operation,
            CASE
              WHEN Period_Edit IS NULL THEN 'Missing Period.'
              WHEN Type_Edit IS NULL THEN 'Missing Type.'
            ELSE
              NULL
            END AS Error,
            'ERROR' AS Severity,
          COALESCE(Line_Item, Insertion_Order, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_Pacing`
        )
        WHERE
          Error IS NOT NULL
      )

      SELECT * FROM INPUT_ERRORS
      ;
    """.format(**project.task),
      legacy=False)

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "PATCH_Pacing",
    """SELECT *
      FROM `{dataset}.SHEET_Pacing`
      WHERE (
        REGEXP_CONTAINS(Insertion_Order, r" - (\d+)$")
        OR REGEXP_CONTAINS(Line_Item, r" - (\d+)$")
      )
      AND Line_Item NOT IN (SELECT Id FROM `{dataset}.AUDIT_Pacing` WHERE Severity='ERROR')
      AND Insertion_Order NOT IN (SELECT Id FROM `{dataset}.AUDIT_Pacing` WHERE Severity='ERROR')
    """.format(**project.task),
    legacy=False
  )


def pacing_patch(commit=False):

  patches = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"PATCH_Pacing",
    }},
    as_object=True
  )

  for row in rows:

    pacing = {}

    if row['Period'] != row['Period_Edit']:
      pacing.setdefault("pacing", {})
      pacing["pacing"]["pacingPeriod"] = row['Period_Edit']

    if row['Type'] != row['Type_Edit']:
      pacing.setdefault("pacing", {})
      pacing["pacing"]["pacingType"] = row['Type_Edit']

    if row['Daily_Budget'] != row['Daily_Budget_Edit']:
      pacing.setdefault("pacing", {})
      pacing["pacing"]["dailyMaxMicros"] = int(float(row['Daily_Budget_Edit']) * 100000)

    if row['Daily_Impressions'] != row['Daily_Impressions_Edit']:
      pacing.setdefault("pacing", {})
      pacing["pacing"]["dailyMaxImpressions"] = row['Daily_Impressions_Edit']

    if pacing:
      patch = {
          "operation": "Pacing",
          "action": "PATCH",
          "partner": row['Partner'],
          "advertiser": row['Advertiser'],
          "campaign": row['Campaign'],
          "parameters": {
              "advertiserId": lookup_id(row['Advertiser']),
              "body": pacing
          }
      }

      if row['Line_Item']:
        patch["line_item"] = row['Line_Item']
        patch["parameters"]["lineItemId"] = lookup_id(row['Line_Item'])
      else:
        patch["insertion_order"] = row['Insertion_Order']
        patch["parameters"]["insertionOrderId"] = lookup_id(row['Insertion_Order'])

      patches.append(patch)

  patch_masks(patches)

  if commit:
    insertion_order_commit(patches)
    line_item_commit(patches)
  else:
    patch_preview(patches)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from django import forms
from django.utils.html import mark_safe

from starthinker_ui.project.models import Project


class ProjectForm(forms.ModelForm):
  share = forms.ChoiceField()

  class Meta:
    model = Project
    fields = ['service', 'key', 'share']

  def __init__(self, account, *args, **kwargs):
    super(ProjectForm, self).__init__(*args, **kwargs)
    self.instance.account = account
    self.fields['service'].label = 'Service JSON'
    self.fields['service'].help_text = mark_safe(
      'Optional, paste service JSON here.'
    )

    self.fields['key'].label = 'API Key'
    self.fields['key'].required = False
    self.fields['key'].help_text = mark_safe(
      'Optional, ensures billing and may be required for some API calls.'
    )

    self.fields['share'].label = 'Share'
    self.fields['share'].help_text = mark_safe(
      'WARNING: Sharing with DOMAIN or GLOBAL will allow other users to use, but not view, your service credentials and project permissions.'
    )

    domain = account.get_domain()
    self.fields['share'].choices = ((
        'user', 'USER | SAFE: Only you can use it.'
    ), ('domain',
        'DOMAIN ( %s ) | CAUTION: Only people in your verified domain can use it.'
        % domain), ('global', 'GLOBAL | CAUTION: Anyone logging in can use it.'
                   )) if domain else (('user', 'User'), ('global', 'Global'))

  def clean_service(self):
    try:
      service = self.cleaned_data['service']
      self.instance.identifier = json.loads(service)['client_email']
    except:
      raise forms.ValidationError('Service credentials must be proper JSON.')

    if Project.objects.filter(
        account=self.instance.account,
        identifier=self.instance.identifier).exclude(
            pk=self.instance.pk).exists():
      raise forms.ValidationError(
          'You already have service credentails for %s, this is a duplicate.' %
          self.instance.identifier)

    return service

  def get_errors(self):
    return self.errors



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
import argparse
import textwrap

from starthinker.util.project import project
from starthinker.util.google_api import API_DCM
from starthinker.util.dcm import get_profile_for_api, report_to_rows, report_clean, report_file, report_schema
from starthinker.util.csv import rows_to_type, rows_print


def main():

  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
      Command line to help debug CM reports and build reporting tools.

      Examples: 
        To get list of reports: python helper.py --account [id] --list -u [user credentials path]
        To get report: python helper.py --account [id] --report [id] -u [user credentials path]
        To get report files: python helper.py --account [id] --files [id] -u [user credentials path]
        To get report sample: python helper.py --account [id] --sample [id] -u [user credentials path]
        To get report schema: python helper.py --account [id] --schema [id] -u [user credentials path]

  """))

  parser.add_argument(
      '--account', help='Account ID to use to pull the report.', default=None)
  parser.add_argument(
      '--report', help='Report ID to pull JSON definition.', default=None)
  parser.add_argument(
      '--schema', help='Report ID to pull achema definition.', default=None)
  parser.add_argument(
      '--sample', help='Report ID to pull sample data.', default=None)
  parser.add_argument(
      '--files', help='Report ID to pull file list.', default=None)
  parser.add_argument('--list', help='List reports.', action='store_true')

  # initialize project
  project.from_commandline(parser=parser, arguments=('-u', '-c', '-s', '-v'))
  auth = 'service' if project.args.service else 'user'

  is_superuser, profile = get_profile_for_api(auth, project.args.account)
  kwargs = {
      'profileId': profile,
      'accountId': project.args.account
  } if is_superuser else {
      'profileId': profile
  }

  # get report list
  if project.args.report:
    kwargs['reportId'] = project.args.report
    report = API_DCM(
        auth, internal=is_superuser).reports().get(**kwargs).execute()
    print(json.dumps(report, indent=2, sort_keys=True))

  # get report files
  elif project.args.files:
    kwargs['reportId'] = project.args.files
    for rf in API_DCM(
        auth, internal=is_superuser).reports().files().list(**kwargs).execute():
      print(json.dumps(rf, indent=2, sort_keys=True))

  # get schema
  elif project.args.schema:
    filename, report = report_file(auth, project.args.account,
                                   project.args.schema, None, 10)
    rows = report_to_rows(report)
    rows = report_clean(rows)
    print(json.dumps(report_schema(next(rows)), indent=2, sort_keys=True))

  # get sample
  elif project.args.sample:
    filename, report = report_file(auth, project.args.account,
                                   project.args.sample, None, 10)
    rows = report_to_rows(report)
    rows = report_clean(rows)
    rows = rows_to_type(rows)
    for r in rows_print(rows, row_min=0, row_max=20):
      pass

  # get list
  else:
    for report in API_DCM(
        auth, internal=is_superuser).reports().list(**kwargs).execute():
      print(json.dumps(report, indent=2, sort_keys=True))


if __name__ == '__main__':
  main()



#https://cloud.google.com/bigquery/docs/share-access-views

from starthinker.util.project import project
from starthinker.util.bigquery import datasets_create, datasets_access, query_to_table, query_to_view


# BUG: can't mix PLX with BigQuery, so this needs to be fixed.
def get_query(task):
  query = """
    SELECT
      SESSION_USER() AS Current_User,
      *
    FROM `%s.%s.%s`
    WHERE
      Advertiser_ID IN (
        SELECT DISTINCT CAST(advertiser_id AS INT64)
        FROM `google.com:dbm-util.plximport.dbm_access_hourly`
        WHERE
          primary_email_hashed = SHA256(LOWER(SESSION_USER()))
            OR
          (
            SELECT has_global_read
            FROM `google.com:dbm-util.plximport.dbm_access_hourly` access
            WHERE
              primary_email_hashed = SHA256(LOWER(SESSION_USER()))
                AND
              has_global_read
            LIMIT 1
          )
      )
  """ % (project.id, project.task['dataset'], project.task['table'])

  #print(query)
  return query


@project.from_parameters
def dbm_authorize():
  if project.verbose:
    print('Authorizing', project.task['dataset'], project.task['table'])

  authorized_dataset = '%s_Authorized_DBM' % project.task['dataset']
  authorized_view = '%s_Authorized_DBM' % project.task['table']

  # Create a separate dataset to store the authorized view
  datasets_create(project.task['auth'], project.id, authorized_dataset)

  # Create the authorized view in the new dataset
  query_to_view(
      project.task['auth'],
      project.id,
      authorized_dataset,
      authorized_view,
      get_query(project.task),
      legacy=False)

  # Assign access controls to the dataset containing the view
  datasets_access(
      project.task['auth'],
      project.id,
      authorized_dataset,
      groups=project.task['groups'],
      role='READER')

  # Authorize the view to access the source dataset
  datasets_access(
      project.task['auth'],
      project.id,
      project.task['dataset'],
      views=[{
          'dataset': authorized_dataset,
          'view': authorized_view
      }])


if __name__ == '__main__':
  dbm_authorize()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.sheets import sheets_create, sheets_tab_create, sheets_read, sheets_write, sheets_clear, sheets_tab_copy, sheets_tab_delete
from starthinker.util.bigquery import get_schema, rows_to_table
from starthinker.util.csv import rows_to_type
from starthinker.util.data import get_rows


@project.from_parameters
def sheets():
  if project.verbose:
    print('SHEETS')

  # if sheet or tab is missing, don't do anything
  if not project.task.get('sheet') or not project.task.get('tab'):
    if project.verbose:
      print('Missing Sheet and/or Tab, skipping task.')
    return

  # delete if specified, will delete sheet if no more tabs remain
  if project.task.get('delete', False):
    sheets_tab_delete(project.task['auth'], project.task['sheet'],
                      project.task['tab'])

  # create a sheet and tab if specified, if template
  if 'template' in project.task:
    sheets_create(
        project.task['auth'],
        project.task['sheet'],
        project.task['tab'],
        project.task['template'].get('sheet'),
        project.task['template'].get('tab'),
    )

  # copy template if specified ( clear in this context means overwrite )
  #if project.task.get('template', {}).get('sheet'):
  #  sheets_tab_copy(
  #    project.task['auth'],
  #    project.task['template']['sheet'],
  #    project.task['template']['tab'],
  #    project.task['sheet'],
  #    project.task['tab'],
  #    project.task.get('clear', False)
  #  )

  # if no template at least create tab
  #else:
  #  sheets_tab_create(
  #    project.task['auth'],
  #    project.task['sheet'],
  #    project.task['tab']
  #  )

  # clear if specified
  if project.task.get('clear', False):
    sheets_clear(project.task['auth'], project.task['sheet'],
                 project.task['tab'], project.task.get('range', 'A1'))

  # write data if specified
  if 'write' in project.task:
    rows = get_rows(project.task['auth'], project.task['write'])
    sheets_write(
        project.task['auth'],
        project.task['sheet'],
        project.task['tab'],
        project.task['range'],
        rows,
        append=False)

  # append data if specified
  if 'append' in project.task:
    rows = get_rows(project.task['auth'], project.task['append'])
    sheets_write(
        project.task['auth'],
        project.task['sheet'],
        project.task['tab'],
        project.task['range'],
        rows,
        append=True)

  # move data if specified
  # move data if specified
  if 'out' in project.task:
    rows = sheets_read(project.task['auth'], project.task['sheet'],
                       project.task['tab'], project.task.get('range', 'A1'))

    if rows:
      schema = None

      # RECOMMENDED: define schema in json
      if project.task['out']['bigquery'].get('schema'):
        if project.verbose:
          print('SHEETS SCHEMA DEFINED')
        schema = project.task['out']['bigquery']['schema']

      # NOT RECOMMENDED: determine schema if missing
      else:
        if project.verbose:
          print(
              'SHEETS SCHEMA DETECT ( Note Recommended - Define Schema In JSON )'
          )
        # cast rows to types ( for schema detection )
        rows = rows_to_type(rows)
        rows, schema = get_schema(
            rows,
            project.task.get('header', False),
            infer_type=project.task.get('infer_type', True))

      # write to table ( not using put because no use cases for other destinations )
      rows_to_table(
          auth=project.task['out'].get('auth', project.task['auth']),
          project_id=project.id,
          dataset_id=project.task['out']['bigquery']['dataset'],
          table_id=project.task['out']['bigquery']['table'],
          rows=rows,
          schema=schema,
          skip_rows=1 if project.task.get('header', False) else 0,
          disposition=project.task['out']['bigquery'].get(
              'disposition', 'WRITE_TRUNCATE'))

    else:
      print('SHEET EMPTY')


if __name__ == '__main__':
  sheets()



# Generated by Django 2.2.13 on 2020-11-19 10:59

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('identifier', models.CharField(db_index=True, max_length=64, unique=True)),
                ('email', models.EmailField(max_length=255, unique=True)),
                ('name', models.CharField(blank=True, default='', max_length=255)),
                ('domain', models.CharField(blank=True, default='', max_length=255)),
                ('picture', models.CharField(blank=True, default='', max_length=255)),
                ('is_active', models.BooleanField(default=True)),
                ('is_admin', models.BooleanField(default=False)),
                ('birthday', models.DateField(auto_now_add=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Command line to schedule recipe execution.

This script is meant to be called form a crontab.  See installdeploy.sh.
It will read a directory and check each *.json recipe for a schedule.  If the
recipe has a task to run during the current time zone adjusted hour, it is
executed.

To add a schedule to any recipe include the following JSON.

  {
    "setup": {
      "timezone": "America/Los_Angeles",
      "day": [ "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun" ],
      "hour": [ "3", "18" ]
    }
  }

To execute each task at a different hour add the following JSON to each task.
Each hour
must match at least one in the the setup section.  Tasks with no hours specified
execute
every hour in setup section.

  {
    "tasks": [
      "sample_task": {
        "hour":[3]
       }
    ]
  }

To run the cron for the current hour:

python cron/run.py [path to recipe directory] [see optional arguments below]

Arguments

  path - run all json files in the specified path
  --project / -p - cloud id of project
  --user / -u - path to user credentials json file, defaults to
  GOOGLE_APPLICATION_CREDENTIALS
  --service / -s - path to service credentials json file
  --client / -c' - path to client credentials json file
  --verbose / -v - print all the steps as they happen.

Each recipe can run under different credentials, specify project, client, user,
and service
values in the JSON for each recipe. See /util/project/README.md.

CAUTION

This script triggers the all/run.py script if the schedule matches the current
hour.
This script does NOT check if the last job finished, potentially causing
overruns.

"""

import sys
import subprocess
import argparse

from glob import glob

from starthinker.config import UI_ROOT

if __name__ == '__main__':

  parser = argparse.ArgumentParser()
  parser.add_argument(
      'path', help='run all json files in the specified path', action='store')

  parser.add_argument(
      '--project',
      '-p',
      help='cloud id of project, defaults to None',
      default=None)
  parser.add_argument(
      '--user',
      '-u',
      help='path to user credentials json file, defaults to GOOGLE_APPLICATION_CREDENTIALS',
      default=None)
  parser.add_argument(
      '--service',
      '-s',
      help='path to service credentials json file, defaults None',
      default=None)
  parser.add_argument(
      '--client',
      '-c',
      help='path to client credentials json file, defaults None',
      default=None)

  parser.add_argument(
      '--verbose',
      '-v',
      help='print all the steps as they happen.',
      action='store_true')
  parser.add_argument(
      '--force',
      '-f',
      help='execute all scripts once then exit.',
      action='store_true')

  args = parser.parse_args()

  for filepath in glob('%s/*.json' % args.path):
    if args.verbose:
      print('RECIPE:', filepath)
    command = 'python -W ignore %s/starthinker/all/run.py %s %s' % (
        UI_ROOT, filepath, ' '.join(sys.argv[2:]))
    if args.verbose:
      print('COMMAND:', command)
    subprocess.Popen(command, shell=True)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.patch import patch_clear
from starthinker.task.dv_editor.patch import patch_log
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def line_item_clear():
  table_create(
      project.task["auth_bigquery"],
      project.id,
      project.task["dataset"],
      "DV_LineItems",
      Discovery_To_BigQuery("displayvideo",
                            "v1").method_schema("advertisers.lineItems.list"),
  )

  sheets_clear(project.task["auth_sheets"], project.task["sheet"], "Line Items",
               "A2:Z")


def line_item_load():

  # load multiple partners from user defined sheet
  def line_item_load_multiple():
    rows = get_rows(
        project.task["auth_sheets"], {
            "sheets": {
                "sheet": project.task["sheet"],
                "tab": "Advertisers",
                "range": "A2:A"
            }
        })

    for row in rows:
      yield from API_DV360(
          project.task["auth_dv"], iterate=True).advertisers().lineItems().list(
              advertiserId=lookup_id(row[0])).execute()

  # write line_items to database
  put_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset":
                  project.task["dataset"],
              "table":
                  "DV_LineItems",
              "schema":
                  Discovery_To_BigQuery(
                      "displayvideo",
                      "v1").method_schema("advertisers.lineItems.list"),
              "format":
                  "JSON"
          }
      }, line_item_load_multiple())

  # write line_items to sheet
  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "query": """SELECT
        CONCAT(P.displayName, ' - ', P.partnerId),
        CONCAT(A.displayName, ' - ', A.advertiserId),
        CONCAT(C.displayName, ' - ', C.campaignId),
        CONCAT(I.displayName, ' - ', I.insertionOrderId),
        CONCAT(L.displayName, ' - ', L.lineItemId),
        'PATCH',
        L.entityStatus,
        ARRAY_TO_STRING(L.warningMessages, '\\n'),

        L.lineItemType,
        L.lineItemType,

        L.flight.flightDateType,
        L.flight.flightDateType,
        CONCAT(L.flight.dateRange.startDate.year, '-', L.flight.dateRange.startDate.month, '-', L.flight.dateRange.startDate.day),
        CONCAT(L.flight.dateRange.startDate.year, '-', L.flight.dateRange.startDate.month, '-', L.flight.dateRange.startDate.day),
        CONCAT(L.flight.dateRange.endDate.year, '-', L.flight.dateRange.endDate.month, '-', L.flight.dateRange.endDate.day),
        CONCAT(L.flight.dateRange.endDate.year, '-', L.flight.dateRange.endDate.month, '-', L.flight.dateRange.endDate.day),
        L.flight.triggerId,
        L.flight.triggerId,

        L.budget.budgetAllocationType,
        L.budget.budgetAllocationType,
        L.budget.budgetUnit,
        L.budget.budgetUnit,
        L.budget.maxAmount / 100000,
        L.budget.maxAmount / 100000,

        L.partnerRevenueModel.markupType,
        L.partnerRevenueModel.markupType,
        CAST(L.partnerRevenueModel.markupAmount AS FLOAT64) / 100000,
        CAST(L.partnerRevenueModel.markupAmount AS FLOAT64) / 100000,

        CAST(L.conversionCounting.postViewCountPercentageMillis AS Float64) / 1000,
        CAST(L.conversionCounting.postViewCountPercentageMillis AS Float64) / 1000,

        L.targetingExpansion.targetingExpansionLevel,
        L.targetingExpansion.targetingExpansionLevel,
        L.targetingExpansion.excludeFirstPartyAudience,
        L.targetingExpansion.excludeFirstPartyAudience,

        FROM `{dataset}.DV_LineItems` AS L
        LEFT JOIN `{dataset}.DV_Advertisers` AS A
        ON L.advertiserId=A.advertiserId
        LEFT JOIN `{dataset}.DV_Campaigns` AS C
        ON L.advertiserId=C.advertiserId
        LEFT JOIN `{dataset}.DV_InsertionOrders` AS I
        ON L.insertionOrderId=I.insertionOrderId
        LEFT JOIN `{dataset}.DV_Partners` AS P
        ON A.partnerId=P.partnerId
      """.format(**project.task),
      "legacy": False
    }}
  )

  put_rows(
    project.task["auth_sheets"],
    { "sheets": {
      "sheet": project.task["sheet"],
      "tab": "Line Items",
      "range": "A2"
    }},
    rows
  )


def line_item_audit():

  rows = get_rows(
    project.task["auth_sheets"],
    { "sheets": {
      "sheet": project.task["sheet"],
      "tab": "Line Items",
      "range": "A2:Z"
    }}
  )

  put_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table": "SHEET_LineItems",
      "schema": [
        { "name": "Partner", "type": "STRING" },
        { "name": "Advertiser", "type": "STRING" },
        { "name": "Campaign", "type": "STRING" },
        { "name": "Insertion_Order", "type": "STRING" },
        { "name": "Line_Item", "type": "STRING" },
        { "name": "Action", "type": "STRING" },
        { "name": "Status", "type": "STRING" },
        { "name": "Warning", "type": "STRING" },
        { "name": "Line_Item_Type", "type": "STRING" },
        { "name": "Line_Item_Type_Edit", "type": "STRING" },
        { "name": "Flight_Data_Type", "type": "STRING" },
        { "name": "Flight_Data_Type_Edit", "type": "STRING" },
        { "name": "Flight_Start_Date", "type": "STRING" },
        { "name": "Flight_Start_Date_Edit", "type": "STRING" },
        { "name": "Flight_End_Date", "type": "STRING" },
        { "name": "Flight_End_Date_Edit", "type": "STRING" },
        { "name": "Flight_Trigger", "type": "STRING" },
        { "name": "Flight_Trigger_Edit", "type": "STRING" },
        { "name": "Budget_Allocation_Type", "type": "STRING" },
        { "name": "Budget_Allocation_Type_Edit", "type": "STRING" },
        { "name": "Budget_Unit", "type": "STRING" },
        { "name": "Budget_Unit_Edit", "type": "STRING" },
        { "name": "Budget_Max", "type": "FLOAT" },
        { "name": "Budget_Max_Edit", "type": "FLOAT" },
        { "name": "Partner_Revenue_Model", "type": "STRING" },
        { "name": "Partner_Revenue_Model_Edit", "type": "STRING" },
        { "name": "Partner_Markup", "type": "FLOAT" },
        { "name": "Partner_Markup_Edit", "type": "FLOAT" },
        { "name": "Conversion_Percent", "type": "FLOAT" },
        { "name": "Conversion_Percent_Edit", "type": "FLOAT" },
        { "name": "Targeting_Expansion_Level", "type": "STRING" },
        { "name": "Targeting_Expansion_Level_Edit", "type": "STRING" },
        { "name": "Exclude_1P", "type": "STRING" },
        { "name": "Exclude_1P_Edit", "type": "STRING" },
      ],
      "format": "CSV"
    }},
    rows
  )

  # Create Insert View
  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "INSERT_LineItems",
    """SELECT
      REGEXP_EXTRACT(S_LI.Advertiser, r' - (\d+)$') AS advertiserId,
      REGEXP_EXTRACT(S_LI.Campaign, r' - (\d+)$') AS campaignId,
      REGEXP_EXTRACT(S_LI.Insertion_Order, r' - (\d+)$') AS insertionOrderId,
      S_LI.Line_Item AS displayName,
      S_LI.Line_Item_Type_Edit AS lineItemType,
      'ENTITY_STATUS_DRAFT' AS entityStatus,
      STRUCT(
        S_PC.Cost_Type_Edit As costType,
        S_PC.Fee_Type_Edit As feeType,
        S_PC.Invoice_Type_Edit AS invoiceType,
        S_PC.Fee_Amount_Edit AS feeAmount,
        S_PC.Fee_Percent_Edit * 1000 AS feePercentageMillis
      ) AS partnerCosts,
      STRUCT(
        S_LI.Flight_Data_Type_Edit AS flightDateType,
        STRUCT (
          STRUCT (
            EXTRACT(YEAR FROM CAST(S_LI.Flight_Start_Date_Edit AS Date)) AS year,
            EXTRACT(MONTH FROM CAST(S_LI.Flight_Start_Date_Edit AS DATE)) AS month,
            EXTRACT(DAY FROM CAST(S_LI.Flight_Start_Date_Edit AS DATE)) AS day
          ) AS startDate,
          STRUCT (
            EXTRACT(YEAR FROM CAST(S_LI.Flight_End_Date_Edit AS Date)) AS year,
            EXTRACT(MONTH FROM CAST(S_LI.Flight_End_Date_Edit AS DATE)) AS month,
            EXTRACT(DAY FROM CAST(S_LI.Flight_End_Date_Edit AS DATE)) AS day
          ) AS endDate
        ) AS dateRange,
        S_LI.Flight_Trigger_Edit AS triggerId
      ) AS flight,
      STRUCT(
        S_LI.Budget_Allocation_Type_Edit AS budgetAllocationType,
        S_LI.Budget_Unit_Edit AS budgetUnit,
        S_LI.Budget_Max_Edit * 100000 AS maxAmount
      ) AS budget,
      STRUCT(
        S_P.Period_Edit As pacingPeriod,
        S_P.Type_Edit As pacingType,
        S_P.Daily_Budget_Edit AS dailyMaxMicros,
        S_P.Daily_Impressions_Edit AS dailyMaxImpressions
      ) AS pacing,
      STRUCT(
        S_FC.Unlimited_Edit AS unlimited,
        S_FC.Time_Unit_Edit AS timeUnit,
        S_FC.Time_Count_Edit AS timeUnitCount,
        S_FC.Max_impressions_Edit AS maxImpressions
      ) AS frequencyCap,
      STRUCT(
        S_LI.Partner_Revenue_Model_Edit AS markupType,
        S_LI.Partner_Markup_Edit * 100000 AS markupAmount
      ) AS partnerRevenueModel,
      STRUCT(
        S_LI. Conversion_Percent_Edit * 1000 AS postViewCountPercentageMillis,
        [] AS floodlightActivityConfigs
      ) AS conversionCounting,
      STRUCT(
        IF(S_BS.Fixed_Bid_Edit IS NOT NULL,
          STRUCT(
            S_BS.Fixed_Bid_Edit * 100000 AS bidAmountMicros
          ),
          NULL
        ) AS fixedBid,
        IF(S_BS.Auto_Bid_Goal_Edit IS NOT NULL,
          STRUCT(
            S_BS.Auto_Bid_Goal_Edit AS performanceGoalType,
            S_BS.Auto_Bid_Amount_Edit * 100000 AS maxAverageCpmBidAmountMicros,
            S_BS.Auto_Bid_Algorithm_Edit AS customBiddingAlgorithmId
          ),
          NULL
        ) AS maximizeSpendAutoBid,
        IF(S_BS.Performance_Goal_Type_Edit IS NOT NULL,
          STRUCT(
            S_BS.Performance_Goal_Type_Edit AS performanceGoalType,
            S_BS.Performance_Goal_Amount_Edit * 100000 AS performanceGoalAmountMicros,
            S_BS.Performance_Goal_Average_CPM_Bid_Edit * 100000 AS maxAverageCpmBidAmountMicros,
            S_BS.Performance_Goal_Algorithm_Edit AS customBiddingAlgorithmId
          ),
          NULL
        ) AS performanceGoalAutoBid
      )
      AS bidStrategy,
      STRUCT(
        S_ID.Integration_Code_Edit AS integrationCode,
        S_ID.Details_Edit AS details
      ) AS integrationDetails,
      STRUCT(
        S_LI.Targeting_Expansion_Level_Edit AS targetingExpansionLevel,
        S_LI.Exclude_1P_Edit AS excludeFirstPartyAudience
      ) AS targetingExpansion
    FROM `{dataset}.SHEET_LineItems` AS S_LI
    LEFT JOIN `{dataset}.SHEET_PartnerCosts` AS S_PC ON S_LI.Line_Item=S_PC.Line_Item
    LEFT JOIN `{dataset}.SHEET_Pacing` AS S_P ON S_LI.Line_Item=S_P.Line_Item
    LEFT JOIN `{dataset}.SHEET_FrequencyCaps` AS S_FC ON S_LI.Line_Item=S_FC.Line_Item
    LEFT JOIN `{dataset}.SHEET_IntegrationDetails` AS S_ID ON S_LI.Line_Item=S_ID.Line_Item
    LEFT JOIN `{dataset}.SHEET_BidStrategy` AS S_BS ON S_LI.Line_Item=S_BS.Line_Item
    LEFT JOIN `{dataset}.DV_LineItems` AS DV_LI ON S_LI.Line_Item=DV_LI.displayName
    WHERE S_LI.Action="INSERT"
    AND DV_LI IS NULL
    """.format(**project.task),
    legacy=False
  )

  # Create Audit View
  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "AUDIT_LineItems",
    """WITH
      /* Check if sheet values are set */
      INPUT_ERRORS AS (
        SELECT
        *
        FROM (
          SELECT
            'Line Item' AS Operation,
            CASE
              WHEN Budget_Allocation_Type_Edit IS NULL THEN 'Missing Budget Allocation Type.'
              WHEN Budget_Unit_Edit IS NULL THEN 'Missing Budget Unit.'
              WHEN Budget_Max_Edit IS NULL THEN 'Missing Budget Max.'
              WHEN Partner_Revenue_Model_Edit IS NULL THEN 'Missing Partner Revenue Model.'
            ELSE
              NULL
            END AS Error,
            'ERROR' AS Severity,
          COALESCE(Line_Item, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_LineItems`
        )
        WHERE
          Error IS NOT NULL
      ),
      /* Check duplicate inserts */
      DUPLICATE_ERRORS AS (
        SELECT
          'Line Item' AS Operation,
          'Duplicate Line Item name, insert will be ignored.' AS Error,
          'WARNING' AS Severity,
          COALESCE(S_LI.Line_Item, 'BLANK') AS Id
        FROM `{dataset}.SHEET_LineItems` As S_LI
        LEFT JOIN `{dataset}.DV_LineItems` AS DV_LI ON S_LI.Line_Item=DV_LI.displayName
        WHERE S_LI.Action="INSERT"
        AND DV_LI IS NOT NULL
      )

      SELECT * FROM INPUT_ERRORS
      UNION ALL
      SELECT * FROM DUPLICATE_ERRORS
    """.format(**project.task),
    legacy=False
  )

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "PATCH_LineItems",
    """SELECT *
      FROM `{dataset}.SHEET_LineItems`
      WHERE Line_Item NOT IN (SELECT Id FROM `{dataset}.AUDIT_LineItems` WHERE Severity='ERROR')
    """.format(**project.task),
    legacy=False
  )


def line_item_patch(commit=False):

  def date_edited(value):
    y, m, d = value.split("-")
    return {"year": y, "month": m, "day": d}

  patches = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"PATCH_LineItems",
    }},
    as_object=True
  )

  for row in rows:

    if row['Action'] == "DELETE":
      patches.append({
        "operation": "Line Items",
        "action": "DELETE",
        "partner": row['Partner'],
        "advertiser": row['Advertiser'],
        "campaign": row['Campaign'],
        "line_item": row['Line_Item'],
        "parameters": {
          "advertiserId": lookup_id(row['Advertiser']),
          "lineItemId": lookup_id(row['Line_Item'])
        }
      })

    elif row['Action'] == "PATCH":
      line_item = {}

      if row['Line_Item_Type'] != row['Line_Item_Type_Edit']:
        line_item["lineItemType"] = row['Line_Item_Type_Edit']

      if row['Flight_Data_Type'] != row['Flight_Data_Type_Edit']:
        line_item.setdefault("flight", {})
        line_item["flight"]["flightDateType"] = row['Flight_Data_Type_Edit']

      if row['Flight_Start_Date'] != row['Flight_Start_Date_Edit']:
        line_item.setdefault("flight", {}).setdefault("dateRange", {})
        line_item["flight"]["dateRange"]["startDate"] = date_edited(row['Flight_Start_Date_Edit'])

      if row['Flight_End_Date'] != row['Flight_End_Date_Edit']:
        line_item.setdefault("flight", {}).setdefault("endDate", {})
        line_item["flight"]["dateRange"]["endDate"] = date_edited(row['Flight_End_Date_Edit'])

      if row['Flight_Trigger'] != row['Flight_Trigger_Edit']:
        line_item.setdefault("flight", {})
        line_item["flight"]["triggerId"] = row['Flight_Trigger_Edit']

      if row['Budget_Allocation_Type'] != row['Budget_Allocation_Type_Edit']:
        line_item.setdefault("budget", {})
        line_item["budget"]["budgetAllocationType"] = row['Budget_Allocation_Type_Edit']

      if row['Budget_Unit'] != row['Budget_Unit_Edit']:
        line_item.setdefault("budget", {})
        line_item["budget"]["budgetUnit"] = row['Budget_Unit_Edit']

      if row['Budget_Max'] != row['Budget_Max_Edit']:
        line_item.setdefault("budget", {})
        line_item["budget"]["maxAmount"] = int(
          float(row['Budget_Max_Edit']) * 100000
        )

      if row['Partner_Revenue_Model'] != row['Partner_Revenue_Model_Edit']:
        line_item.setdefault("partnerRevenueModel", {})
        line_item["partnerRevenueModel"]["markupType"] = row['Partner_Revenue_Model_Edit']

      if row['Partner_Markup'] != row['Partner_Markup_Edit']:
        line_item.setdefault("partnerRevenueModel", {})
        line_item["partnerRevenueModel"]["markupAmount"] = int(
          float(row['Partner_Markup_Edit']) * 100000
        )

      if row['Conversion_Percent'] != row['Conversion_Percent_Edit']:
        line_item.setdefault("conversionCounting", {})
        line_item["conversionCounting"]["postViewCountPercentageMillis"] = int(
          float(row['Conversion_Percent_Edit']) * 1000
        )

      if row['Targeting_Expansion_Level'] != row['Targeting_Expansion_Level_Edit']:
        line_item.setdefault("targetingExpansion", {})
        line_item["targetingExpansion"]["targetingExpansionLevel"] = row['Targeting_Expansion_Level_Edit']

      if row['Exclude_1P'] != row['Exclude_1P_Edit']:
        line_item.setdefault("targetingExpansion", {})
        line_item["targetingExpansion"]["excludeFirstPartyAudience"] = row['Exclude_1P_Edit']

      if line_item:
        patches.append({
          "operation": "Line Items",
          "action": "PATCH",
          "partner": row['Partner'],
          "advertiser": row['Advertiser'],
          "campaign": row['Campaign'],
          "line_item": row['Line_Item'],
          "parameters": {
            "advertiserId": lookup_id(row['Advertiser']),
            "lineItemId": lookup_id(row['Line_Item']),
            "body": line_item
          }
        })

  patch_masks(patches)

  if commit:
    line_item_commit(patches)
  else:
    patch_preview(patches)


def line_item_insert(commit=False):
  inserts = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"INSERT_LineItems",
    }},
    as_object=True
  )

  for row in rows:
    inserts.append({
      "operation": "Line Items",
      "action": "INSERT",
      "partner": None,
      "advertiser": row['advertiserId'],
      "campaign": row['campaignId'],
      "line_item": row['displayName'],
      "parameters": {
        "advertiserId": row['advertiserId'],
        "body":row
      }
    })

  if commit:
    line_item_commit(inserts)
  else:
    patch_preview(inserts)


def line_item_commit(patches):
  for patch in patches:
    if not patch.get("line_item"):
      continue
    print("API LINE ITEM:", patch["action"], patch["line_item"])
    try:
      if patch["action"] == "DELETE":
        response = API_DV360(
          project.task["auth_dv"]
        ).advertisers().lineItems().delete(
          **patch["parameters"]
        ).execute()
        patch["success"] = response
      elif patch["action"] == "PATCH":
        response = API_DV360(
          project.task["auth_dv"]
        ).advertisers().lineItems().patch(
          **patch["parameters"]
        ).execute()
        patch["success"] = response["lineItemId"]
      elif patch["action"] == "INSERT":
        response = API_DV360(
          project.task["auth_dv"]
        ).advertisers().lineItems().create(
          **patch["parameters"]
        ).execute()
        patch["success"] = response["lineItemId"]
    except Exception as e:
      patch["error"] = str(e)
    finally:
      patch_log(patch)
  patch_log()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from datetime import datetime

from starthinker.util.project import project
from starthinker.util.dbm import report_file, report_get, report_clean, report_to_rows, DBM_CHUNKSIZE, report_to_list
from starthinker.util.bigquery import rows_to_table, make_schema
from starthinker.util.sdf import get_single_sdf_rows
from starthinker.util.data import put_rows

BUDGET_SEGMENTS = 'Budget Segments'
IO_ID = 'Io Id'
CHANGES_SCHEMA = ['Io Id', 'Category', 'Old Value', 'Delta', 'New Value']


@project.from_parameters
def monthly_budget_mover():
  if project.verbose:
    print('MONTHLY BUDGET MOVER')

  # Get Spend report
  r = report_get(project.task['auth'], name=project.task['report_name'])

  report_id = report_get(
      project.task['auth'], name=project.task['report_name'])['queryId']
  #	report = report_to_list(project.task['auth'], report_id)

  # Get Insertion Order SDF
  # sdf = list(get_single_sdf_rows(
  #     project.task['auth'],
  #     project.task['sdf']['version'],
  #     project.task['sdf']['partner_id'],
  #     project.task['sdf']['file_types'],
  #     project.task['sdf']['filter_type'],
  #     project.task['sdf']['read']['filter_ids'],
  #    	'InsertionOrders'))

  # print('sdf============================================================')
  # print(sdf)
  # print('sdf============================================================')

  sdf = [
      [
          'Io Id', 'Campaign Id', 'Name', 'Timestamp', 'Status', 'Io Type',
          'Billable Outcome', 'Fees', 'Integration Code', 'Details', 'Pacing',
          'Pacing Rate', 'Pacing Amount', 'Frequency Enabled',
          'Frequency Exposures', 'Frequency Period', 'Frequency Amount',
          'Performance Goal Type', 'Performance Goal Value', 'Measure DAR',
          'Measure DAR Channel', 'Budget Type', 'Budget Segments',
          'Auto Budget Allocation', 'Geography Targeting - Include',
          'Geography Targeting - Exclude', 'Language Targeting - Include',
          'Language Targeting - Exclude', 'Device Targeting - Include',
          'Device Targeting - Exclude', 'Browser Targeting - Include',
          'Browser Targeting - Exclude', 'Digital Content Labels - Exclude',
          'Brand Safety Sensitivity Setting', 'Brand Safety Custom Settings',
          'Third Party Verification Services',
          'Third Party Verification Labels', 'Channel Targeting - Include',
          'Channel Targeting - Exclude', 'Site Targeting - Include',
          'Site Targeting - Exclude', 'App Targeting - Include',
          'App Targeting - Exclude', 'App Collection Targeting - Include',
          'App Collection Targeting - Exclude', 'Category Targeting - Include',
          'Category Targeting - Exclude', 'Keyword Targeting - Include',
          'Keyword Targeting - Exclude', 'Keyword List Targeting - Exclude',
          'Audience Targeting - Similar Audiences',
          'Audience Targeting - Include', 'Audience Targeting - Exclude',
          'Affinity & In Market Targeting - Include',
          'Affinity & In Market Targeting - Exclude', 'Custom List Targeting',
          'Inventory Source Targeting - Authorized Seller Options',
          'Inventory Source Targeting - Include',
          'Inventory Source Targeting - Exclude',
          'Inventory Source Targeting - Target New Exchanges',
          'Daypart Targeting', 'Daypart Targeting Time Zone',
          'Environment Targeting', 'Viewability Targeting Active View',
          'Position Targeting - Display On Screen',
          'Position Targeting - Video On Screen',
          'Position Targeting - Display Position In Content',
          'Position Targeting - Video Position In Content',
          'Position Targeting - Audio Position In Content',
          'Video Player Size Targeting', 'Demographic Targeting Gender',
          'Demographic Targeting Age', 'Demographic Targeting Household Income',
          'Demographic Targeting Parental Status', 'Connection Speed Targeting',
          'Carrier Targeting - Include', 'Carrier Targeting - Exclude',
          'Insertion Order Optimization', 'Bid Strategy Unit',
          'Bid Strategy Do Not Exceed', 'Apply Floor Price For Deals',
          'Algorithm Id'
      ],
      [
          '3489402', '294948', 'Audit_creas_Mustang',
          '2020-03-06T22:04:11.821000', 'Paused', 'Standard', 'Impression',
          '(Media; 0.0; Display & Video 360 Fee; True;);', '', '', 'Flight',
          'Even', '0', 'True', '1', 'Lifetime', '0', 'None', '0', 'False', '',
          'Amount',
          '(12000.0; 05/01/2020; 05/31/2020;);(12000.0; 06/01/2020; 06/30/2020;);',
          'False', '', '', '', '', '2; 502; 202; 302;', '', '', '', '',
          'Use custom',
          'Adult; Alcohol; Derogatory; Downloads & Sharing; Drugs; Gambling; '
          'Profanity; Religion; Sensitive social issues; Suggestive; Tobacco; '
          'Tragedy; Transportation Accidents; Violence; Weapons;',
          'None', '', '',
          '2203109; 2998109; 2998110; 2998111; 2998112; 2998113; 2998114;', '',
          '', '', '', '', '', '', '', '', '', '', 'False', '', '', '', '', '',
          'Authorized and Non-Participating Publisher',
          '1; 6; 8; 9; 10; 2; 11; 12; 13; 16; 17; 19; 20; 21; 23; 27; 31; 34; '
          '36; 37; 38; 41; 42; 43; 50; 52; 60;',
          '', 'True', '', '', 'Web; App;', '', '', '', '', '', '', '', '', '',
          '', '', '', '', '', 'False', '', '', '', ''
      ],
      [
          '3502002', '294948', 'FR_Mindshare_Ford_Mustang_CTX_VOL_Avril17',
          '2020-03-06T12:07:04.366000', 'Paused', 'Standard', 'Impression',
          '(Media; 0.0; Display & Video 360 Fee; True;);', '', '', 'Flight',
          'Even', '0', 'False', '0', 'Minutes', '0', 'None', '0', 'False', '',
          'Amount',
          '(12000.0; 05/01/2020; 05/31/2020;);(12000.0; 06/01/2020; 06/30/2020;);',
          'False', '2250;', '', '', '', '2;', '502; 202; 302;', '', '', '',
          'Use custom',
          'Adult; Alcohol; Derogatory; Downloads & Sharing; Drugs; Gambling; '
          'Politics; Profanity; Religion; Sensitive social issues; Suggestive;'
          ' Tobacco; Tragedy; Transportation Accidents; Violence; Weapons;',
          'None', '', '',
          '2203109; 2998109; 2998110; 2998111; 2998112; 2998113; 2998114;', '',
          '', '', '', '', '', '', '', '', '', '', 'False', '', '', '', '', '',
          'Authorized and Non-Participating Publisher',
          '1; 6; 8; 9; 10; 2; 11; 12; 13; 16; 17; 19; 20; 21; 23; 27; 31; 34; '
          '36; 37; 38; 41; 42; 43; 50; 52; 60;',
          '', 'True', '', '', 'Web; App;', '', '', '', '', '', '', '', '', '',
          '', '', '', '', '', 'False', '', '', '', ''
      ],
      [
          '3522675', '294948', 'FR_Mindshare_Ford_Sales_Juin_CTX_VOL_Juin17',
          '2020-03-06T10:44:48.709000', 'Paused', 'Standard', 'Impression',
          '(Media; 0.0; Display & Video 360 Fee; True;);', '', '', 'Flight',
          'Ahead', '0', 'False', '0', 'Minutes', '0', 'None', '0', 'False', '',
          'Amount',
          '(12000.0; 05/01/2020; 05/31/2020;);(12000.0; 06/01/2020; 06/30/2020;);',
          'False', '2250;', '', '', '', '2;', '502; 202; 302;', '', '', '',
          'Use custom',
          'Adult; Alcohol; Derogatory; Downloads & Sharing; Drugs; Gambling; '
          'Profanity; Religion; Sensitive social issues; Suggestive; Tobacco; '
          'Tragedy; Transportation Accidents; Violence; Weapons;',
          'None', '', '',
          '2203109; 2998109; 2998110; 2998111; 2998112; 2998113; 2998114;', '',
          '', '', '', '', '', '', '', '', '', '', 'False', '', '', '', '', '',
          'Authorized and Non-Participating Publisher',
          '1; 6; 8; 9; 10; 2; 11; 12; 13; 16; 17; 19; 20; 21; 23; 27; 31; 34; '
          '36; 37; 38; 41; 42; 43; 50; 52; 60;',
          '', 'True', '', '', 'Web; App;', '', '', '', '', '', '', '', '', '',
          '', '', '', '', '', 'False', '', '', '', ''
      ]
  ]

  report = [[
      'Advertiser_Currency', 'Insertion_Order_Id', 'Revenue_Adv_Currency'
  ], ['EUR', '3489402', '893.195881'], ['EUR', '3502002', '14893.195881'],
            ['EUR', '3522675', '893.195881']]
  print('report============================================================')
  print(report)
  print('report============================================================')

  # Prep out blocks depending on where the outputs should be stored
  if project.task['is_colab']:
    project.task['out_old_sdf'].pop('bigquery')
    project.task['out_new_sdf'].pop('bigquery')
    project.task['out_changes'].pop('bigquery')

  else:
    project.task['out_old_sdf'].pop('file')
    project.task['out_new_sdf'].pop('file')
    project.task['out_changes'].pop('file')

    # Build Schemas
    schema = make_schema(sdf[0])
    schema_changes = make_schema(CHANGES_SCHEMA)
    project.task['out_old_sdf']['bigquery']['schema'] = schema
    project.task['out_new_sdf']['bigquery']['schema'] = schema
    project.task['out_changes']['bigquery']['schema'] = schema_changes

  # Write old sdf to table
  put_rows(project.task['auth'], project.task['out_old_sdf'], (n for n in sdf))

  # Categorize the IOs to be aggregated together
  if (project.task['budget_categories'] != {} and
      project.task['budget_categories'] != '' and
      project.task['budget_categories'] != None):

    categories = remove_excluded_ios_from_categories(
        project.task['budget_categories'], project.task['excluded_ios'])

    categories_spend = aggregate_io_spend_to_categories(report, categories)

    categories_budget = aggregate_io_budget_to_categories(sdf, categories)

    category_budget_deltas = calc_budget_spend_deltas(categories_budget,
                                                      categories_spend,
                                                      categories)

    new_sdf, changes = apply_category_budgets(sdf, category_budget_deltas,
                                              categories)

  # Don't split up the IOs by categories
  else:
    report_dict = convert_report_to_dict(report)
    new_sdf, changes = calc_new_sdf_no_categories(sdf, report_dict,
                                                  project.task['excluded_ios'])

  if project.task['is_colab']:
    changes.insert(0, CHANGES_SCHEMA)

  # Write new sdf to table
  put_rows(project.task['auth'], project.task['out_new_sdf'],
           (n for n in new_sdf))

  # Write log file to table
  put_rows(project.task['auth'], project.task['out_changes'],
           (n for n in changes))


""" Convert the spend report into an dictionary: key=>io_id, value=>spend

Args: * report=> DV360 report containing spend data for every IO in the
categories list

Returns:
  * a dictionary: key=>io_id, value=>spend for the previous month
"""


def convert_report_to_dict(report):
  spend_dict = {}

  # Read header to get correct columns then remove header
  report_spend_column = report[0].index('Revenue_Adv_Currency')
  report_io_id_column = report[0].index('Insertion_Order_Id')
  report.pop(0)

  for row in report:
    spend_dict[row[report_io_id_column]] = row[report_spend_column]

  return spend_dict


""" Create new SDF when no categories are provided

Args: * report dictionary=> DV360 report dictionary with io_id and spend amount

Returns:
  * sdf -> updated budget segments
  *changes -> what sort of changes happened to the budget
"""


def calc_new_sdf_no_categories(sdf, report, excluded_ios):
  changes = []
  new_sdf = []

  # Read sdf header for values
  sdf_io_id_column = sdf[0].index(IO_ID)
  sdf_budget_segments_column = sdf[0].index(BUDGET_SEGMENTS)
  is_first = True

  for row in sdf:
    if is_first:
      new_sdf.append(row)
      is_first = False
      continue

    # If Io is in exclude then just add the current
    io_id = row[sdf_io_id_column]
    if io_id in excluded_ios:
      new_sdf.append(row)
      continue

    budget_segment = convert_budget_segment_to_obj(
        row[sdf_budget_segments_column])
    current_month_idx = get_current_month_idx_in_budget_segment(
        budget_segment, str(io_id))

    # Calculate the budget delta from the previous month
    prev_spend = 0
    if io_id in report:
      prev_spend = report[io_id]

    prev_budget = get_prev_month_budget(row[sdf_budget_segments_column], io_id)

    prev_delta = float(prev_budget) - float(prev_spend)

    old_budget = float(budget_segment[current_month_idx][0])
    new_budget = old_budget + prev_delta
    budget_segment[current_month_idx][0] = '{:0.2f}'.format(new_budget)

    # Insert new budget segment information into the row
    new_budget_segment = convert_budget_segment_obj_to_string(budget_segment)
    row[sdf_budget_segments_column] = new_budget_segment

    # Log the change to be written to BQ
    changes.append([io_id, '', old_budget, prev_delta, new_budget])

    new_sdf.append(row)

  return new_sdf, changes


""" Aggregates all the io spend into their categories

Args: * report=> DV360 report containing spend data for every IO in the
categories list * categories => dictionary with categories as the key, and a
list of ios under that category as the value

Returns:
  * a dictionary with the category as the key, and the aggregated spend for the
  IOs in that category as the value
"""


def aggregate_io_spend_to_categories(report, categories):
  processed_ios = {}
  categories_spend = {}

  # Read header to get correct columns and remove header
  spend_column = report[0].index('Revenue (Adv Currency)')
  io_id_column = report[0].index('Insertion Order ID')
  report.pop(0)

  # Aggregate io spend into bigger categories
  for row in report:
    io_id = row[io_id_column]
    io_spend = row[spend_column]

    for key, val in categories.iteritems():
      if val == None:
        continue
      if int(io_id) in val:
        # Add io to the list of processed ios
        if processed_ios.has_key(key):
          processed_ios[key].append(io_id)
        else:
          processed_ios[key] = [io_id]

        # Aggregate category spend
        if categories_spend.has_key(key):
          categories_spend[key] = categories_spend[key] + float(io_spend)
        else:
          categories_spend[key] = float(io_spend)

        break

  validate_all_ios_processed(categories, processed_ios, 'Spend Report')

  return categories_spend


""" Aggregate io budget up to the category level

Args: * sdf => the current sdf for the desired filter ids * categories =>
dictionary with categories as the key, and a list of ios under that category as
the value

Returns:
  * dictionary with the category as the key, and the aggregated budget for that
  category as the value
"""


def aggregate_io_budget_to_categories(sdf, categories):
  processed_ios = {}
  categories_budget = {}

  # Read header to get correct columns and remove header
  header = sdf[0]
  io_id_idx = header.index(IO_ID)
  budget_segments_idx = header.index(BUDGET_SEGMENTS)

  # Aggregate io budget into bigger categories
  for row in sdf:
    if sdf.index(row) == 0:
      continue
    io_id = row[io_id_idx]
    budget_segments = row[budget_segments_idx]

    for key, val in categories.iteritems():
      if int(io_id) in val:
        # Get budget amount from budget segment
        budget = get_prev_month_budget(budget_segments, io_id)

        # Add IO to the list of processed ios
        if processed_ios.has_key(key):
          processed_ios[key].append(io_id)
        else:
          processed_ios[key] = [io_id]

        # Aggregate category budget
        if categories_budget.has_key(key):
          categories_budget[key] = categories_budget[key] + budget
        else:
          categories_budget[key] = budget

  validate_all_ios_processed(categories, processed_ios, 'SDF')

  return categories_budget


""" Removes the ios from the excluded list from the categories list

Args: * categories => dictionary with categories as the key, and a list of ios
under that category as the value * excluded_ios => list of ios to be excluded
from the calculations

Returns:
  * The new dictionary of categories with the necessary ios excluded from the
  list

"""


def remove_excluded_ios_from_categories(categories, excluded_ios):

  for io in excluded_ios:
    for key, value in categories.iteritems():
      if io in value:
        categories[key].remove(io)
        break

  return categories


""" Apply the new calculated budget deltas to the sdf under the budget segments portion

Args: * sdf => the current sdf file to be editted * category_budget_deltas => a
dictionary with category as the key, and the value of the budget - spend for
that category as the value * categories => dictionary with categories as the
key, and a list of ios under that category as the value

Returns:
  * new_sdf => the new sdf in list format with the updated budget segment
  strings inputted
        * changes => a list of changes that occured including: Io_Id, Category,
        Old_Value, Delta, New_Value
"""


def apply_category_budgets(sdf, category_budget_deltas, categories):
  changes = []
  new_sdf = []
  first = True
  io_id_idx = -1
  budget_segments_idx = -1

  for row in sdf:
    # Do not process the header and get the idx for Io Id and
    if first:
      io_id_idx = row.index(IO_ID)
      budget_segments_idx = row.index(BUDGET_SEGMENTS)
      if io_id_idx < 0:
        raise Exception('IO Id column was not found in SDF.')
      if budget_segments_idx < 0:
        raise Exception('Budget Segment column was not found in SDF.')
      new_sdf.append(row)
      first = False
      continue

    # Update budget segments with the new values
    io_id = row[io_id_idx]
    budget_segment = row[budget_segments_idx]
    category = None

    # figure out which category the current IO is in
    for key, val in categories.iteritems():
      if int(io_id) in val:
        category = key

    # Only update the row if the IO has been applied to a category
    if category != None:
      # Get the current budget to be updated
      budget_segment_obj = convert_budget_segment_to_obj(budget_segment)
      current_month_idx = get_current_month_idx_in_budget_segment(
          budget_segment_obj, str(io_id))

      # Apply the budget delta to the budget segment object
      if category in category_budget_deltas and current_month_idx != None:
        old_budget = float(budget_segment_obj[current_month_idx][0])
        budget_segment_obj[current_month_idx][0] = '{:0.2f}'.format(
            old_budget + category_budget_deltas[category])

        # Insert new budget segment information into the row
        new_budget_segment = convert_budget_segment_obj_to_string(
            budget_segment_obj)
        row[budget_segments_idx] = new_budget_segment

        # Log the change to be written to BQ
        changes.append([
            io_id, category, old_budget, category_budget_deltas[category],
            budget_segment_obj[current_month_idx][0]
        ])

    new_sdf.append(row)

  return new_sdf, changes


""" Converts the list of budget segment objects to the string format needed by the SDF

Args: * budget_segment_obj => A list of budget segment objects. [{budget,
start_date, end_date},{...}]

Returns:
  * A string representing the budget segment in the format required by sdfs.
        (#; MM/DD/YYYY; MM/DD/YYYY;); (#; MM/DD/YYYY; MM/DD/YYYY;);
"""


def convert_budget_segment_obj_to_string(budget_segment_obj):
  budget_segment_str = ''
  for segment in budget_segment_obj:
    budget_segment_str += '(' + str(
        segment[0]) + '; ' + segment[1] + '; ' + segment[2] + ';); '

  return budget_segment_str


""" Find the index for the current months budget in the budget segments obj

Args:
  * budget_obj: list of budget_objs

Returns:
  * Index of the current month's budget information in the budget obj
"""


def get_current_month_idx_in_budget_segment(budget_objs, io_id):
  # Seperate budget_string into a list of budget objs
  today = datetime.today()
  month = str(today.month)
  if len(month) == 1:
    month = '0' + month

  # Find correct month in the budget_obj
  for monthly_budget in budget_objs:
    if monthly_budget[1][:2] == month and monthly_budget[1][-4:] == str(
        today.year):
      return budget_objs.index(monthly_budget)

  raise Exception('The current monthly budget is not available ' + month +
                  '. Io Id => ' + io_id)


""" Calculate the deltas between the category budget and spend amount

Args: * categories_budget => dictionary with the category as the key, and the
aggregated budget for that category as the value * categories_spend =>
dictionary with the category as the key, and the aggregated spend from the
previous month as the * categories => dictionary with categories as the key, and
a list of ios under that category as the value

Returns:
  * new_sdf => the new sdf in list format with the updated budget segment
  strings inputted
        * changes => a list of changes that occured including: Io_Id, Category,
        Old_Value, Delta, New_Value
"""


def calc_budget_spend_deltas(categories_budget, categories_spend, categories):
  budget_deltas = {}

  for key, budget in categories_budget.iteritems():
    delta = budget - categories_spend[key]
    budget_deltas[key] = delta / len(categories[key])

  return budget_deltas


""" Gets the budget allocated for the previous month

Args: * budget_string => the budget segment string from the sdf (#; MM/DD/YYYY;
MM/DD/YYYY;); (#; MM/DD/YYYY; MM/DD/YYYY;);

Returns:
  * a float that represents the budget for the previous month
"""


def get_prev_month_budget(budget_string, io_id):
  # Seperate budget_string into a list of budget objs
  budget_obj = convert_budget_segment_to_obj(budget_string)

  # Calculate previous month, with two characters
  cur_year = str(datetime.today().year)
  cur_month = datetime.today().month
  prev_month_int = int(cur_month) - 1
  if (prev_month_int == 0):
    prev_month_int = 12
    cur_year = str(int(cur_year) - 1)
  prev_month = str(prev_month_int)
  if (len(prev_month) == 1):
    prev_month = '0' + prev_month

  # Find correct month in the budget_obj
  for monthly_budget in budget_obj:
    if monthly_budget[1][:2] == prev_month and monthly_budget[1][
        -4:] == cur_year:
      return float(monthly_budget[0])

  raise Exception(
      io_id +
      ' does not have budget set for the previous month(month number = ' +
      prev_month + ')')


""" Helper to split the budget segment string from the sdf into a list of budget objects

Args: * budget_string => string from the SDF that represents the budget segment
Format of the string =>

Returns:
  * List of budget objects => [{budget, start_date, end_date},...]
        budget_string format => "(#; MM/DD/YYYY; MM/DD/YYYY;); (#; MM/DD/YYYY;
        MM/DD/YYYY;);"

"""


def convert_budget_segment_to_obj(budget_string):
  budget_obj = []

  one_month_per_item = budget_string.replace('(', '').replace(' ',
                                                              '').split(';);')

  for val in one_month_per_item:
    budget_obj.append(val.split(';'))

  return budget_obj[:-1]


""" Go through the list of ios that were found from the source and verify there are no missing ios from the categories list

Args: * categories => dictionary with categories as the key, and a list of ios
under that category as the value * processed_ios => a dictionary with a list of
ios that were found in the source and aggregated * source => where the
aggregation was happening => Spend Report or SDF

Returns:
  * Nothing, but throws an exception if any un processed ios are found in the
  categories list
"""


def validate_all_ios_processed(categories, processed_ios, source):
  not_processed = []

  for key, value in categories.iteritems():
    for io_id in value:
      if str(io_id) not in processed_ios[key]:
        not_processed.append(io_id)

  if len(not_processed) > 0:
    not_processed_str = ''
    for io in not_processed:
      not_processed_str += str(io) + ', '

    raise Exception('The following ios were not found in the ' + source +
                    ':  ' + not_processed_str)


if __name__ == '__main__':
  monthly_budget_mover()



# Copyright 2011 Google Inc. All Rights Reserved.
"""A thread-compatible singleton decorator.

Sometimes, only a single instance of a class should ever be created. This file
provides a wrapper that turns a class into a Singleton. The constructor may
only be called once; a static method Singleton() provides access to the
constructed instance. If Singleton() is called before the constructor, or if
the constructor is called multiple times, Errors are raised.

The only thread-safety guarantees this wrapper provides are:

* Multiple calls to Singleton() may occur concurrently, and no two calls will
  return different objects.
* `__init__` will never be called twice, even in the face of concurrent calls.
  Exactly one attempt will succeed, all others will raise an Error.

All calls to Singleton() must happen after the constructor call -- so either
the constructor must be invoked before any threads would try to access the
singleton (e.g. in main()), or else the threads must use an additional lock
to synchronize attempts to invoke the constructor. In that case, using a
memoized factory function or a non-singleton global guarded by a lock may be
more appropriate.

Singletons are often associated with bad coding practices; see
http://wiki/Main/SingletonsConsideredDangerous and decide if you should
really be using this functionality.

To make your singletons more testable, use the following idiom:


class A(...):
  "All the complicated code goes in here, can be tested normally..."
  ..
  ..


@singleton.Singleton
class B(A):
  "Singleton instance of A"


Example usage:

from google3.pyglib import singleton


@singleton.Singleton
class Foo(object):
  "Example singleton"


a = Foo()
b = Foo.Singleton()
c = Foo.Singleton()
assert a == b
assert b == c
"""

__author__ = 'elliottk@google.com (Elliott Karpilovsky)'

import threading

_CLASS_LOCKS = {}  # Holds the per-class locks.
_CLASS_LOCKS_LOCK = threading.Lock()  # Lock for obtaining a per-class lock.
_INSTANCES = {}  # Mapping from class to instantiated object.


class Error(Exception):
  """Base error class."""


class AlreadyHasSingletonMethodError(Error):
  """Raised if the class already defines a Singleton() method."""

  def __init__(self, cls):
    Error.__init__(self)
    self.cls = cls

  def __str__(self):
    return 'Class already has a Singleton() method: %s' % self.cls


class NotConstructedError(Error):
  """Raised if the constructor has not been called yet."""

  def __init__(self, cls):
    Error.__init__(self)
    self.cls = cls

  def __str__(self):
    return 'Constructor has not yet been called for class %s' % self.cls


class ConstructorCalledAgainError(Error):
  """Raised if the constructor is called twice for a singleton."""

  def __init__(self, cls, args, kws):
    Error.__init__(self)
    self.cls = cls
    self.args = args
    self.kws = kws

  def __str__(self):
    return ('Constructor called (again) on class %s with args %s and kws %s' %
            (self.cls, self.args, self.kws))


def _GetClassLock(cls):
  """Returns the lock associated with the class."""
  with _CLASS_LOCKS_LOCK:
    if cls not in _CLASS_LOCKS:
      _CLASS_LOCKS[cls] = threading.Lock()
    return _CLASS_LOCKS[cls]


def Singleton(cls):
  """Turn a class into a singleton.

  Exactly one call to the constructor is required. After that, all future calls
  must be to the Singleton() static method.

  This code is multithread-safe. Shared locks are held for brief periods
  of time; when a class is instantiated, it uses a class specific lock.

  Args:
    cls: The class to decorate.

  Returns:
    The singleton class. Note that this class is a sub-class
    of the class it is decorating.

  Raises:
    AlreadyHasSingletonMethodError: If the class has a Singleton method
      defined.
  """
  if hasattr(cls, 'Singleton'):
    raise AlreadyHasSingletonMethodError(cls)

  class _Singleton(cls):

    def __init__(self, *args, **kws):
      class_lock = _GetClassLock(cls)
      with class_lock:
        if cls not in _INSTANCES:
          cls.__init__(self, *args, **kws)
          _INSTANCES[cls] = self
        else:
          raise ConstructorCalledAgainError(cls, args, kws)

    @staticmethod
    def Singleton():
      class_lock = _GetClassLock(cls)
      with class_lock:
        if cls not in _INSTANCES:
          raise NotConstructedError(cls)
        else:
          return _INSTANCES[cls]

  return _Singleton
